# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio, Gst, GObject

from time import sleep
from re import match
from random import shuffle
import json
import os
import tempfile

from lollypop.logger import Logger
from lollypop.utils import escape, emit_signal
from lollypop.define import App, Type
from lollypop.objects_track import Track
from lollypop.objects_album import Album


class MtpSyncDb:
    """
        Synchronisation db stored on MTP device

        Some MTP devices for instance cannot properly store / retrieve file
        modification times, so we store them in a dedicated file at the root
        of the MTP device instead.

        The storage format is a simple JSON dump.
        It also implements the context manager interface, ensuring database is
        loaded before entering the scope and saving it when exiting.
    """

    def __init__(self):
        """
            Constructor for MtpSyncDb
        """
        self.__encoder = "convert_none"
        self.__normalize = False
        self.__metadata = {}

    def load(self, base_uri):
        """
            Loads the metadata db from the MTP device
            @param base_uri as str
        """
        self.__base_uri = base_uri
        self.__db_uri = self.__base_uri + "/lollypop-sync.db"
        Logger.debug("MtpSyncDb::__load_db()")
        try:
            dbfile = Gio.File.new_for_uri(self.__db_uri)
            (status, jsonraw, tags) = dbfile.load_contents(None)
            if status:
                jsondb = json.loads(jsonraw.decode("utf-8"))
                if "encoder" in jsondb:
                    self.__encoder = jsondb["encoder"]
                if "normalize" in jsondb:
                    self.__normalize = jsondb["normalize"]
                if "version" in jsondb and jsondb["version"] == 1:
                    for m in jsondb["tracks_metadata"]:
                        self.__metadata[m["uri"]] = m["metadata"]
                else:
                    Logger.info("MtpSyncDb::__load_db():"
                                " unknown sync db version")
        except Exception as e:
            Logger.error("MtpSyncDb::load(): %s" % e)

    def save(self):
        """
            Saves the metadata db to the MTP device
        """
        try:
            Logger.debug("MtpSyncDb::__save()")
            jsondb = json.dumps(
                            {"version": 1,
                             "encoder": self.__encoder,
                             "normalize": self.__normalize,
                             "tracks_metadata": [
                                 {"uri": x, "metadata": y}
                                 for x, y in sorted(self.__metadata.items())]})
            dbfile = Gio.File.new_for_uri(self.__db_uri)
            (tmpfile, stream) = Gio.File.new_tmp()
            stream.get_output_stream().write_all(jsondb.encode("utf-8"))
            tmpfile.copy(dbfile, Gio.FileCopyFlags.OVERWRITE, None, None)
            stream.close()
            tmpfile.delete(None)
        except Exception as e:
            Logger.error("MtpSyncDb::__save(): %s", e)

    def set_encoder(self, encoder):
        """
            Set encoder
            @param encoder as str
        """
        self.__encoder = encoder

    def set_normalize(self, normalize):
        """
            Set normalize
            @param normalize as bool
        """
        self.__normalize = normalize

    def get_mtime(self, uri):
        """
            Get mtime for a uri on MTP device from the metadata db
            @param uri as str
        """
        return self.__metadata.get(
            self.__get_reluri(uri), {}).get("time::modified", 0)

    def set_mtime(self, uri, mtime):
        """
            Set mtime for a uri on MTP device from the metadata db
            @param uri as str
            @param mtime as int
        """
        self.__metadata.setdefault(self.__get_reluri(uri),
                                   dict())["time::modified"] = mtime

    def delete_uri(self, uri):
        """
            Deletes metadata for a uri from the on-device metadata db
            @param uri as str
        """
        if self.__get_reluri(uri) in self.__metadata:
            del self.__metadata[self.__get_reluri(uri)]

    @property
    def encoder(self):
        """
            Get encoder
            @return str
        """
        return self.__encoder

    @property
    def normalize(self):
        """
            Get normalize
            @return bool
        """
        return self.__normalize

############
# Private  #
############
    def __get_reluri(self, uri):
        """
            Returns a relative on-device uri from an absolute on-device.
            We do not want to store absolute uri in the db as the same
            peripheral could have a different path when mounted on another host
            machine.
            @param uri as str
        """
        if uri.startswith(self.__base_uri):
            uri = uri[len(self.__base_uri) + 1:]
        return uri


class MtpSync(GObject.Object):
    """
        Synchronisation to MTP devices
    """
    __gsignals__ = {
        "sync-progress": (GObject.SignalFlags.RUN_FIRST, None, (float,)),
        "sync-finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "sync-errors": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    __ENCODE_START = 'filesrc location="%s" ! decodebin\
                            ! audioconvert\
                            ! audioresample\
                            ! audio/x-raw,rate=44100,channels=2'
    __ENCODE_END = ' ! filesink location="%s"'
    __NORMALIZE = " ! rgvolume pre-amp=6.0 headroom=10.0\
                    ! rglimiter ! audioconvert"
    __EXTENSION = {"convert_none": None,
                   "convert_mp3": ".mp3",
                   "convert_vorbis": ".ogg",
                   "convert_flac": ".flac",
                   "convert_aac": ".m4a"}
    __ENCODERS = {"convert_none": None,
                  "convert_mp3": " ! lamemp3enc target=bitrate\
                                   cbr=true bitrate=%s ! id3v2mux",
                  "convert_vorbis": " ! vorbisenc bitrate=%s\
                                      ! oggmux",
                  "convert_flac": " ! flacenc",
                  "convert_aac": " ! faac bitrate=%s ! mp4mux"}
    _GST_ENCODER = {"convert_mp3": ["lamemp3enc", "id3v2mux"],
                    "convert_ogg": ["vorbisenc", "oggmux"],
                    "convert_flac": ["flacenc"],
                    "convert_aac": ["faac", "mp4mux"]}

    def __init__(self):
        """
            Init MTP synchronisation
        """
        GObject.Object.__init__(self)
        self.__cancellable = Gio.Cancellable()
        self.__errors_count = 0
        self.__on_mtp_files = []
        self.__last_error = ""
        self.__uri = None
        self.__total = 0  # Total files to sync
        self.__done = 0   # Handled files on sync
        self.__mtp_syncdb = MtpSyncDb()

    def check_encoder_status(self, encoder):
        """
            Check encoder status
            @param encoder as str
            @return bool
        """
        ok = True
        for item in self._GST_ENCODER[encoder]:
            if not Gst.ElementFactory.find(item):
                ok = False
        return ok

    def sync(self, uri, index):
        """
            Sync playlists with device. If playlists contains Type.NONE,
            sync albums marked as to be synced
            @param uri as str
            @param index as int => device index
        """
        try:
            self.__cancellable = Gio.Cancellable()
            self.__uri = uri
            self.__convert_bitrate = App().settings.get_value(
                "convert-bitrate").get_int32()
            self.__errors_count = 0
            self.__total = 0
            self.__done = 0
            tracks = []

            Logger.info("Getting tracks to sync")
            # New tracks for synced albums
            album_ids = App().albums.get_synced_ids(0)
            album_ids += App().albums.get_synced_ids(index)
            for album_id in album_ids:
                album = Album(album_id, skipped=False)
                tracks += album.tracks
            # New tracks for playlists
            playlist_ids = App().playlists.get_synced_ids(0)
            playlist_ids += App().playlists.get_synced_ids(index)
            for playlist_id in playlist_ids:
                if App().playlists.get_smart(playlist_id):
                    request = App().playlists.get_smart_sql(playlist_id)
                    for track_id in App().db.execute(request):
                        tracks.append(Track(track_id))
                else:
                    for track_id in App().playlists.get_track_ids(playlist_id):
                        tracks.append(Track(track_id))

            Logger.info("Getting URIs to copy")
            uris = self.__get_uris_to_copy(tracks)
            shuffle(uris)
            self.__total = len(uris) + 2

            Logger.info("Deleting old files")
            if not self.__cancellable.is_cancelled():
                self.__delete_old_uris(uris)

            Logger.info("Copying files")
            for (src_uri, dst_uri) in uris:
                try:
                    if self.__cancellable.is_cancelled():
                        break
                    self.__copy_file(src_uri, dst_uri)
                    self.__done += 1
                    emit_signal(self, "sync-progress",
                                self.__done / self.__total)
                except Exception as e:
                    Logger.error("MtpSync::__copy_files(): %s", e)
            Logger.debug("Writing playlists")
            if not self.__cancellable.is_cancelled():
                self.__write_playlists(playlist_ids)
            emit_signal(self, "sync-progress",
                        self.__done / self.__total + 1)
            Logger.debug("Creating unsync")
            if not self.__cancellable.is_cancelled():
                d = Gio.File.new_for_uri(self.__uri + "/unsync")
                if not d.query_exists():
                    d.make_directory_with_parents()
            emit_signal(self, "sync-progress",
                        self.__done / self.__total + 2)
        except Exception as e:
            Logger.error("MtpSync::sync(): %s" % e)
        finally:
            Logger.info("Save sync db")
            if not self.__cancellable.is_cancelled():
                self.__mtp_syncdb.save()
            self.cancel()
            if self.__errors_count != 0:
                Logger.debug("Sync errors")
                emit_signal(self, "sync-errors", self.__last_error)
            Logger.debug("Sync finished")
            emit_signal(self, "sync-finished")

    def cancel(self):
        """
            Cancel sync
        """
        Logger.info("MtpSync::cancel()")
        self.__cancellable.cancel()

    @property
    def db(self):
        """
            Get sync db
        """
        return self.__mtp_syncdb

############
# PRIVATE  #
############
    def __get_album_name(self, track):
        """
            Get on device URI for album
            @param track as Track
            @return URI as str
        """
        album_name = track.album_name.lower()
        is_compilation = track.album.artist_ids[0] == Type.COMPILATIONS
        if is_compilation:
            string = escape(album_name)
        else:
            artists = ", ".join(track.album.artists).lower()
            string = escape("%s_%s" % (artists, album_name))
        return string[:100]

    def __get_uris_to_copy(self, tracks):
        """
            Get on device URI for all tracks
            @param tracks as [Track]
        """
        uris = []
        art_uris = []
        for track in tracks:
            f = Gio.File.new_for_uri(track.uri)
            album_device_uri = "%s/%s" % (self.__uri,
                                            GLib.uri_escape_string(
                                                self.__get_album_name(track),
                                                None,
                                                True))
            album_local_uri = f.get_parent().get_uri()
            src_uri = "%s/%s" % (album_local_uri,
                                 GLib.uri_escape_string(f.get_basename(),
                                                        None,
                                                        True))
            dst_uri = "%s/%s" % (album_device_uri, escape(f.get_basename()))
            (convertion_needed,
             dst_uri) = self.__is_convertion_needed(src_uri, dst_uri)
            uris.append((src_uri, dst_uri))
            art_uri = App().album_art.get_uri(track.album)
            if art_uri is not None:
                art_filename = Gio.File.new_for_uri(art_uri).get_basename()
                art_uris.append((art_uri,
                                 "%s/%s" % (album_device_uri,
                                            escape(art_filename))))
        return uris + art_uris

    def __write_playlists(self, playlist_ids):
        """
            Write playlists on disk
            @param playlist_ids as [int]
        """
        for playlist_id in playlist_ids:
            if self.__cancellable.is_cancelled():
                break
            try:
                # Get tracks
                if App().playlists.get_smart(playlist_id):
                    request = App().playlists.get_smart_sql(playlist_id)
                    track_ids = App().db.execute(request)
                else:
                    track_ids = App().playlists.get_track_ids(playlist_id)

                # Build tracklist
                tracklist = "#EXTM3U\n"
                for track_id in track_ids:
                    if self.__cancellable.is_cancelled():
                        break
                    track = Track(track_id)
                    f = Gio.File.new_for_uri(track.uri)
                    filename = f.get_basename()
                    album_uri = self.__get_album_name(track)
                    uri = "%s/%s" % (album_uri, escape(filename))
                    (convertion_needed,
                     uri) = self.__is_convertion_needed(track.uri, uri)
                    tracklist += "%s\n" % uri

                # Write playlist file
                playlist_name = escape(App().playlists.get_name(playlist_id))
                playlist_uri = "%s/%s.m3u" % (self.__uri, playlist_name)
                Logger.debug("MtpSync::__write_playlists(): %s" % playlist_uri)
                temp_uri = os.path.join(
                    tempfile.gettempdir(),
                    "lollypop_%s.m3u" % playlist_name
                )
                m3u_temp = Gio.File.new_for_path(temp_uri)
                m3u_temp.replace_contents(
                                    tracklist.encode("utf-8"),
                                    None,
                                    False,
                                    Gio.FileCreateFlags.REPLACE_DESTINATION,
                                    self.__cancellable)
                m3u = Gio.File.new_for_uri(playlist_uri)
                m3u_temp.move(m3u, Gio.FileCopyFlags.OVERWRITE, None, None)
            except Exception as e:
                Logger.error("MtpSync::__write_playlists(): %s", e)

    def __delete_old_uris(self, uris):
        """
            Delete old URIs from device
            @param uris as [str]
        """
        on_device_uris = self.__on_device_uris()
        for (src_uri, dst_uri) in uris:
            dst = Gio.File.new_for_uri(dst_uri)
            if dst.get_uri() in on_device_uris:
                on_device_uris.remove(dst.get_uri())
        for uri in on_device_uris:
            if self.__cancellable.is_cancelled():
                break
            try:
                f = Gio.File.new_for_uri(uri)
                parent = f.get_parent()
                f.delete(self.__cancellable)
                self.__mtp_syncdb.delete_uri(uri)
                infos = parent.enumerate_children(
                    "standard::name,standard::type",
                    Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                    None)
                if len(list(infos)) == 0:
                    parent.delete(self.__cancellable)
                infos.close(None)
            except Exception as e:
                Logger.error("MtpSync::__delete_old_uris(): %s", e)

    def __on_device_uris(self):
        """
            Get URIS on device
            @return [str]
        """
        children = []
        dir_uris = [self.__uri]
        d = Gio.File.new_for_uri(self.__uri)
        if not d.query_exists():
            d.make_directory_with_parents(None)
        while dir_uris:
            if self.__cancellable.is_cancelled():
                break
            try:
                uri = dir_uris.pop(0)
                d = Gio.File.new_for_uri(uri)
                infos = d.enumerate_children(
                    "standard::name,standard::type",
                    Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                    None)
                for info in infos:
                    if self.__cancellable.is_cancelled():
                        break
                    if info.get_file_type() == Gio.FileType.DIRECTORY:
                        if info.get_name() != "unsync":
                            f = infos.get_child(info)
                            dir_uris.append(f.get_uri())
                    else:
                        if info.get_name() == "lollypop-sync.db":
                            continue
                        f = infos.get_child(info)
                        children.append(f.get_uri())
                infos.close(None)
            except Exception as e:
                Logger.error("MtpSync::__get_track_files(): %s, %s" % (e, uri))
        return children

    def __is_convertion_needed(self, src_uri, dst_uri):
        """
            Check if convertion is needed
            @param src_uri as str
            @param dst_uri as str
            @return (bool, str) => Is needed and new dst URI
        """
        # Check extension for convertion
        m = match(r".*(\.[^.]*)", src_uri)
        ext = m.group(1)
        convert_ext = self.__EXTENSION[self.__mtp_syncdb.encoder]
        if ext.lower() in [".png", ".jpg", ".jpeg", "gif"]:
            convertion_needed = False
        elif (convert_ext is not None and
                ext != convert_ext) or self.__mtp_syncdb.normalize:
            convertion_needed = True
            dst_uri = dst_uri.replace(ext, convert_ext)
        else:
            convertion_needed = False
        return (convertion_needed, dst_uri)

    def __copy_file(self, src_uri, dst_uri):
        """
            Copy source to destination
            @param src_uri as str
            @param dst_uri as str
        """
        src = Gio.File.new_for_uri(src_uri)
        (convertion_needed,
         dst_uri) = self.__is_convertion_needed(src_uri, dst_uri)
        dst = Gio.File.new_for_uri(dst_uri)
        parent = dst.get_parent()
        if not parent.query_exists():
            parent.make_directory_with_parents()
        info = src.query_info("time::modified",
                              Gio.FileQueryInfoFlags.NONE,
                              None)
        mtime = info.get_attribute_uint64("time::modified")
        if not dst.query_exists() or\
                self.__mtp_syncdb.get_mtime(dst_uri) < mtime:
            Logger.debug("MtpSync::__copy_file(): %s -> %s"
                         % (src_uri, dst_uri))
            if convertion_needed:
                path = "{}/lollypop_convert".format(GLib.get_tmp_dir())
                convert_file = Gio.File.new_for_path(path)
                pipeline = self.__convert(src, convert_file)
                # Check if encoding is finished
                if pipeline is not None:
                    bus = pipeline.get_bus()
                    bus.add_signal_watch()
                    bus.connect("message::eos", self.__on_bus_eos)
                    self.__encoding = True
                    while self.__encoding and\
                            not self.__cancellable.is_cancelled():
                        sleep(1)
                    bus.disconnect_by_func(self.__on_bus_eos)
                    bus.remove_signal_watch()
                    pipeline.set_state(Gst.State.NULL)
                    convert_file.move(
                        dst, Gio.FileCopyFlags.OVERWRITE, None, None)
                    # To be sure
                    try:
                        convert_file.delete(None)
                    except:
                        pass
            else:
                src.copy(dst, Gio.FileCopyFlags.OVERWRITE, None, None)
            self.__mtp_syncdb.set_mtime(dst_uri, mtime)

    def __convert(self, src, dst):
        """
            Convert file to mp3
            @param src as Gio.File
            @param dst as Gio.File
            @return Gst.Pipeline
        """
        if src.get_path() is None:
            Logger.error("Can't convert files over sftp, smb, ...")
            return None
        try:
            # We need to escape \ in path
            src_path = src.get_path().replace("\\", "\\\\\\")
            dst_path = dst.get_path().replace("\\", "\\\\\\")
            pipeline_str = self.__ENCODE_START % src_path
            if self.__mtp_syncdb.normalize:
                pipeline_str += self.__NORMALIZE
            if self.__mtp_syncdb.encoder in ["convert_vorbis", "convert_aac"]:
                convert_bitrate = self.__convert_bitrate * 1000
            else:
                convert_bitrate = self.__convert_bitrate
            try:
                pipeline_str += self.__ENCODERS[self.__mtp_syncdb.encoder] %\
                    convert_bitrate
            except:
                pipeline_str += self.__ENCODERS[self.__mtp_syncdb.encoder]
            pipeline_str += self.__ENCODE_END % dst_path
            pipeline = Gst.parse_launch(pipeline_str)
            pipeline.set_state(Gst.State.PLAYING)
            return pipeline
        except Exception as e:
            Logger.error("MtpSync::__convert(): %s" % e)
            return None

    def __on_bus_eos(self, bus, message):
        """
            Stop encoding
            @param bus as Gst.Bus
            @param message as Gst.Message
        """
        self.__encoding = False



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, GLib, Gtk, Gdk

import cairo

from lollypop.define import App, ArtBehaviour
from lollypop.utils import get_round_surface


class ArtHelper(GObject.Object):
    """
        Helper to load artwork smoothly
    """

    def __init__(self):
        """
            Init helper
        """
        GObject.Object.__init__(self)

    def set_frame(self, image, frame, width, height):
        """
            Update image for frame
            @param image as Gtk.Image
            @param frame as str
            @param width as int
            @param height as int
        """
        context = image.get_style_context()
        for c in context.list_classes():
            context.remove_class(c)
        context.add_class(frame)
        padding = context.get_padding(Gtk.StateFlags.NORMAL)
        border = context.get_border(Gtk.StateFlags.NORMAL)
        image.set_size_request(width + padding.left +
                               padding.right + border.left + border.right,
                               height + padding.top +
                               padding.bottom + border.top + border.bottom)

    def set_album_artwork(self, album, width, height, scale_factor,
                          effect, callback, *args):
        """
            Set artwork for album id
            @param album as Album
            @param width as int
            @param height as int
            @param scale_factor as int
            @param effect as ArtBehaviour
            @param callback as function
        """
        App().task_helper.run(App().album_art.get,
                              album,
                              width,
                              height,
                              scale_factor,
                              effect,
                              callback=(self._on_get_artwork_pixbuf,
                                        width,
                                        height,
                                        scale_factor,
                                        effect,
                                        callback,
                                        *args))

    def set_artist_artwork(self, name, width, height, scale_factor,
                           effect, callback, *args):
        """
            Set artwork for album id
            @param name as str
            @param width as int
            @param height as int
            @param scale_factor as int
            @param effect as ArtBehaviour
            @param callback as function
        """
        App().task_helper.run(App().artist_art.get,
                              name,
                              width,
                              height,
                              scale_factor,
                              effect,
                              callback=(self._on_get_artwork_pixbuf,
                                        width,
                                        height,
                                        scale_factor,
                                        effect,
                                        callback,
                                        *args))

#######################
# PROTECTED           #
#######################
    def _on_get_artwork_pixbuf(self, pixbuf, width, height, scale_factor,
                               effect, callback, *args):
        """
            Transform pixbuf to surface and load surface effects
            @param pixbuf as Gdk.Pixbuf
            @param size as int
            @param scale_factor as int
            @param effect as ArtBehaviour
            @param callback as function
        """
        surface = None
        if pixbuf is not None:
            if effect & ArtBehaviour.ROUNDED:
                radius = pixbuf.get_width() / 2
                surface = get_round_surface(pixbuf, scale_factor, radius)
            elif effect & ArtBehaviour.ROUNDED_BORDER:
                surface = get_round_surface(pixbuf, scale_factor, 5)
            else:
                surface = Gdk.cairo_surface_create_from_pixbuf(
                        pixbuf, scale_factor, None)
        App().task_helper.run(self.__surface_effects, surface, width, height,
                              scale_factor, effect, callback, *args)

#######################
# PRIVATE             #
#######################
    def __surface_effects(self, surface, width, height, scale_factor,
                          effect, callback, *args):
        """
            Load surface effects
            @param surface as cairo.Surface
            @param size as int
            @param scale_factor as int
            @param effect as ArtBehaviour
            @param callback as function
        """
        if effect & ArtBehaviour.DARKER:
            self.__set_color(surface, 0, 0, 0)
        if effect & ArtBehaviour.LIGHTER:
            self.__set_color(surface, 1, 1, 1)
        GLib.idle_add(callback, surface, *args)

    def __set_color(self, surface, r, g, b):
        """
            Get a darker pixbuf
            @param surface as cairo.Surface
            @param r as int
            @param g as int
            @param b as int
            @return cairo.Surface
        """
        if surface is not None:
            ctx = cairo.Context(surface)
            ctx.rectangle(0, 0, surface.get_width(), surface.get_height())
            ctx.set_source_rgba(r, g, b, 0.5)
            ctx.fill()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio

from gettext import gettext as _
from time import time

from lollypop.objects_track import Track
from lollypop.define import App
from lollypop.logger import Logger


class RatingWidget(Gtk.Bin):
    """
        Rate widget
    """

    def __init__(self, object, icon_size=Gtk.IconSize.BUTTON):
        """
            Init widget
            @param object as Track/Album
            @param is album as bool
            @param icon_size as Gtk.IconSize
        """
        Gtk.Bin.__init__(self)
        self.__object = object
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/RatingWidget.ui")
        builder.connect_signals(self)
        self.__empty_star = builder.get_object("empty_star")
        self._stars = []
        for i in range(0, 5):
            star = builder.get_object("star%s" % i)
            star.set_from_icon_name("starred-symbolic", icon_size)
            self._stars.append(star)
        self.__empty_star.set_from_icon_name("starred-symbolic", icon_size)
        self._on_leave_notify_event(None, None)
        self.add(builder.get_object("widget"))
        if isinstance(object, Track):
            play_count = object.popularity
            self.set_tooltip_text(_("Song played %s times") % play_count)

    def set_icon_size(self, icon_size):
        """
            Set widget icon size
            @param icon_size as Gtk.IconSize
        """
        for start in self._stars:
            start.set_from_icon_name("starred-symbolic", icon_size)
        self.__empty_star.set_from_icon_name("starred-symbolic", icon_size)

#######################
# PROTECTED           #
#######################
    def _on_enter_notify_event(self, widget, event):
        """
            Update star opacity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        event_star = widget.get_children()[0]
        # First star is hidden (used to clear score)
        if event_star.get_opacity() == 0.0:
            found = True
        else:
            found = False
        for star in self._stars:
            if found:
                star.set_opacity(0.2)
                star.get_style_context().remove_class("selected")
            else:
                star.get_style_context().add_class("selected")
                star.set_opacity(0.8)
            if star == event_star:
                found = True

    def _on_leave_notify_event(self, widget, event):
        """
            Update star opacity
            @param widget as Gtk.EventBox (can be None)
            @param event as Gdk.Event (can be None)
        """
        user_rating = True
        rate = self.__object.rate
        if rate < 1:
            rate = self.__object.get_popularity()
            user_rating = False
            for i in range(5):
                self._stars[i].set_opacity(0.2)
                self._stars[i].get_style_context().remove_class("selected")
        if rate > 0:
            star = self.__star_from_rate(rate)
            # Select wanted star
            for idx in range(0, star):
                widget = self._stars[idx]
                if user_rating:
                    widget.get_style_context().add_class("selected")
                else:
                    widget.get_style_context().remove_class("selected")
                widget.set_opacity(0.8)
            # Unselect others
            for idx in range(star, 5):
                self._stars[idx].set_opacity(0.2)
                self._stars[idx].get_style_context().remove_class("selected")

    def _on_button_release_event(self, widget, event):
        """
            Set album popularity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        user_rating = True
        rate = self.__object.rate
        if rate < 1:
            rate = self.__object.get_popularity()
            user_rating = False
        max_star = self.__star_from_rate(rate)
        event_star = widget.get_children()[0]
        if event_star in self._stars:
            position = self._stars.index(event_star)
        else:
            position = -1
        pop = position + 1
        if event.button != 1:
            self.__object.set_popularity(pop)
            self.__object.set_rate(0)
            self._on_leave_notify_event(None, None)
        elif pop == 0 or pop == max_star:
            if user_rating:
                self.__object.set_rate(0)
            else:
                self.__object.set_popularity(0)
            self._on_leave_notify_event(None, None)
        else:
            self.__object.set_rate(pop)

        # Save to tags if needed
        if App().settings.get_value("save-to-tags") and\
                isinstance(self.__object, Track) and\
                self.__object.id >= 0:
            App().task_helper.run(self.__set_popularity, pop)
        return True

#######################
# PRIVATE             #
#######################
    def __star_from_rate(self, rate):
        """
            Calculate stars from rate
            @param rate as double
            @return int
        """
        star = min(5, int(rate))
        return star

    def __set_popularity(self, pop):
        """
            Set popularity with kid3-cli
            @param pop as int
        """
        if pop == 0:
            value = 0
        elif pop == 1:
            value = 1
        elif pop == 2:
            value = 64
        elif pop == 3:
            value = 128
        elif pop == 4:
            value = 196
        else:
            value = 255
        f = Gio.File.new_for_uri(self.__object.uri)
        if f.query_exists():
            path = f.get_path()
            arguments = [["kid3-cli", "-c", "set POPM %s" % value, path],
                         ["flatpak-spawn", "--host", "kid3-cli", "-c",
                          "set POPM %s" % value, path]]
            if App().scanner.inotify is not None:
                App().scanner.inotify.disable()
            worked = False
            for argv in arguments:
                try:
                    (pid, stdin, stdout, stderr) = GLib.spawn_async(
                        argv, flags=GLib.SpawnFlags.SEARCH_PATH |
                        GLib.SpawnFlags.STDOUT_TO_DEV_NULL,
                        standard_input=False,
                        standard_output=False,
                        standard_error=False
                    )
                    GLib.spawn_close_pid(pid)
                    # Force mtime update to not run a collection update
                    App().tracks.set_mtime(self.__object.id, int(time()) + 10)
                    worked = True
                    break
                except Exception as e:
                    Logger.error(
                        "RatingWidget::__on_can_set_popularity(): %s" % e)
            if not worked:
                App().notify.send("Lollypop",
                                  _("You need to install kid3-cli"))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, GObject

from lollypop.utils import emit_signal


class Popover(Gtk.Popover):
    """
        Overlay to be compatible with MenuWidget
    """
    # Same signal than MenuWidget
    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self, auto_destroy=True):
        """
            Init widget
            @param auto_destroy as bool
        """
        Gtk.Popover.__init__(self)
        self.connect("closed", self.__on_closed)
        self.__auto_destroy = auto_destroy

    def __on_closed(self, popover):
        """
            Destroy self
        """
        emit_signal(self, "hidden", True)
        if self.__auto_destroy:
            GLib.idle_add(self.destroy)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from locale import strcoll

from lollypop.view_flowbox import FlowBoxView
from lollypop.define import App, Type, ViewType, StorageType
from lollypop.utils import popup_widget
from lollypop.utils_album import tracks_to_albums
from lollypop.objects_track import Track
from lollypop.widgets_playlist_rounded import PlaylistRoundedWidget
from lollypop.widgets_banner_playlists import PlaylistsBannerWidget
from lollypop.shown import ShownPlaylists
from lollypop.helper_signals import SignalsHelper, signals_map


class PlaylistsManagerView(FlowBoxView, SignalsHelper):
    """
        Show playlists in a FlowBox
    """

    @signals_map
    def __init__(self, view_type):
        """
            Init decade view
            @param view_type as ViewType
        """
        FlowBoxView.__init__(self, StorageType.ALL,
                             view_type | ViewType.OVERLAY)
        self.__signal_id = None
        self._empty_icon_name = "folder-documents-symbolic"
        if self.args is None:
            self.add_widget(self._box, None)
        else:
            self.__banner = PlaylistsBannerWidget(self)
            self.__banner.show()
            self.add_widget(self._box, self.__banner)
        return [
            (App().playlists, "playlists-added", "_on_playlist_added"),
            (App().playlists, "playlists-removed", "_on_playlist_removed"),
            (App().playlists, "playlists-renamed", "_on_playlist_renamed")
        ]

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            FlowBoxView.populate(self, items)

        def load():
            items = [i[0] for i in ShownPlaylists.get()]
            items += App().playlists.get_ids()
            return items

        App().task_helper.run(load, callback=(on_load,))

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"view_type": self.view_type}

#######################
# PROTECTED           #
#######################
    def _get_child(self, value):
        """
            Get a child for view
            @param value as object
            @return row as SelectionListRow
        """
        if self.destroyed:
            return None
        # Compatibility with SelectionList
        if isinstance(value, tuple):
            value = value[0]
        widget = PlaylistRoundedWidget(value, self.view_type,
                                       self.font_height)
        self._box.insert(widget, -1)
        widget.show()
        return widget

    def _sort_func(self, widget1, widget2):
        """
            Sort function
            @param widget1 as PlaylistRoundedWidget
            @param widget2 as PlaylistRoundedWidget
        """
        # Static vs static
        if widget1.data < 0 and widget2.data < 0:
            return widget1.data < widget2.data
        # Static entries always on top
        elif widget2.data < 0:
            return True
        # Static entries always on top
        if widget1.data < 0:
            return False
        # String comparaison for non static
        else:
            return strcoll(widget1.name, widget2.name)

    def _on_child_activated(self, flowbox, child):
        """
            Navigate into child
            @param flowbox as Gtk.FlowBox
            @param child as Gtk.FlowBoxChild
        """
        App().window.container.show_view([Type.PLAYLISTS], child.data)

    def _on_tertiary_press_gesture(self, x, y, event):
        """
            Play artist
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        child = self._box.get_child_at_pos(x, y)
        if child is None or child.artwork is None:
            return
        track_ids = []
        if child.data > 0 and App().playlists.get_smart(child.data):
            request = App().playlists.get_smart_sql(child.data)
            if request is not None:
                track_ids = App().db.execute(request)
        else:
            track_ids = App().playlists.get_track_ids(child.data)
        tracks = [Track(track_id) for track_id in track_ids]
        albums = tracks_to_albums(tracks)
        if albums:
            App().player.play_album_for_albums(albums[0], albums)

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Show Context view for activated album
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        self._on_primary_long_press_gesture(x, y)

    def _on_primary_long_press_gesture(self, x, y):
        """
            Show Context view for activated album
            @param x as int
            @param y as int
        """
        child = self._box.get_child_at_pos(x, y)
        if child is None or child.artwork is None:
            return
        self.__popup_menu(child)

    def _on_playlist_added(self, playlists, playlist_id):
        """
            Add playlist
            @param playlists as Playlists
            @param playlist_id as int
        """
        self.add_value(playlist_id)

    def _on_playlist_removed(self, playlists, playlist_id):
        """
            Remove playlist
            @param playlists as Playlists
            @param playlist_id as int
        """
        for child in self._box.get_children():
            if child.data == playlist_id:
                child.destroy()
                break

    def _on_playlist_renamed(self, playlists, playlist_id):
        """
            Rename playlist
            @param playlists as Playlists
            @param playlist_id as int
        """
        item = None
        for child in self._box.get_children():
            if child.data == playlist_id:
                item = child
                break
        if item is not None:
            name = App().playlists.get_name(playlist_id)
            item.rename(name)

#######################
# PRIVATE             #
#######################
    def __popup_menu(self, child):
        """
            Popup menu for playlist
            @param child as PlaylistRoundedWidget
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_playlist import PlaylistMenu, PlaylistMenuExt
        menu = PlaylistMenu(child.data, self.view_type,
                            App().window.folded)
        menu_widget = MenuBuilder(menu)
        if child.data >= 0:
            menu_widget = MenuBuilder(menu)
            menu_ext = PlaylistMenuExt(child.data)
            menu_ext.show()
            menu_widget.add_widget(menu_ext)
        else:
            menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, child, None, None, None)


class PlaylistsManagerDeviceView(PlaylistsManagerView):
    """
        Show playlists in a FlowBox
    """

    def __init__(self, index, view_type=ViewType.SCROLLED):
        """
            Init decade view
            @param index as int
            @param view_type as ViewType
        """
        PlaylistsManagerView.__init__(self, view_type)
        self.__index = index

    def populate(self):
        """
            Populate items
            @param items
        """
        def on_load(items):
            FlowBoxView.populate(self, items)

        def load():
            items = App().playlists.get_synced_ids(0)
            items += App().playlists.get_synced_ids(self.__index)
            return items

        App().task_helper.run(load, callback=(on_load,))

    @property
    def args(self):
        return None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version("Gst", "1.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GstAudio", "1.0")
gi.require_version("GstPbutils", "1.0")
gi.require_version("TotemPlParser", "1.0")
gi.require_version("Handy", "1")
from gi.repository import Gtk, Gio, GLib, Gdk, Gst, GstPbutils, Handy
Gst.init(None)
GstPbutils.pb_utils_init()

from threading import current_thread
from pickle import dump
from signal import signal, SIGINT, SIGTERM

from lollypop.utils import init_proxy_from_gnome
from lollypop.application_actions import ApplicationActions
from lollypop.application_cmdline import ApplicationCmdline
from lollypop.utils_file import install_youtube_dl
from lollypop.define import LOLLYPOP_DATA_PATH, StorageType
from lollypop.database import Database
from lollypop.player import Player
from lollypop.inhibitor import Inhibitor
from lollypop.artwork import Artwork
from lollypop.artwork_album import AlbumArtwork
from lollypop.artwork_artist import ArtistArtwork
from lollypop.logger import Logger
from lollypop.ws_director import DirectorWebService
from lollypop.sqlcursor import SqlCursor
from lollypop.settings import Settings
from lollypop.database_cache import CacheDatabase
from lollypop.database_albums import AlbumsDatabase
from lollypop.database_artists import ArtistsDatabase
from lollypop.database_genres import GenresDatabase
from lollypop.database_tracks import TracksDatabase
from lollypop.notification import NotificationManager
from lollypop.playlists import Playlists
from lollypop.helper_task import TaskHelper
from lollypop.helper_art import ArtHelper
from lollypop.collection_scanner import CollectionScanner


class Application(Gtk.Application, ApplicationActions, ApplicationCmdline):
    """
        Lollypop application:
            - Handle appmenu
            - Handle command line
            - Create main window
    """

    def __init__(self, version, data_dir, app_id):
        """
            Create application
            @param version as str
            @param data_dir as str
            @param app_id as str
        """
        Gtk.Application.__init__(
            self,
            application_id=app_id,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)
        ApplicationCmdline.__init__(self, version)
        self.__app_id = app_id
        self.__data_dir = data_dir
        self.set_property("register-session", True)
        signal(SIGINT, lambda a, b: self.quit())
        signal(SIGTERM, lambda a, b: self.quit())
        # Set main thread name
        # We force it to current python 3.6 name, to be sure in case of
        # change in python
        current_thread().setName("MainThread")
        (self.__proxy_host, self.__proxy_port) = init_proxy_from_gnome()
        GLib.setenv("PULSE_PROP_media.role", "music", True)
        GLib.setenv("PULSE_PROP_application.icon_name",
                    app_id, True)
        # Flatpak hacks
        if GLib.file_test("/app", GLib.FileTest.EXISTS):
            # Set /tmp for GLib, /tmp not accessible in flatpak
            tmp = GLib.environ_getenv(GLib.get_environ(), "XDG_RUNTIME_DIR")
            GLib.setenv("TMPDIR", "%s/app/%s" % (tmp, app_id), True)
        self.cursors = {}
        self.shown_sidebar_tooltip = False
        self.system_supports_color_schemes = False
        self.__window = None
        self.__fs_window = None
        settings = Gio.Settings.new("org.gnome.desktop.interface")
        self.animations = settings.get_value("enable-animations").get_boolean()
        GLib.set_application_name("Lollypop")
        GLib.set_prgname("org.gnome.Lollypop")
        self.connect("activate", self.__on_activate)
        self.connect("shutdown", lambda a: self.__save_state())
        if GLib.environ_getenv(GLib.get_environ(), "DEBUG_LEAK") is not None:
            import gc
            gc.set_debug(gc.DEBUG_LEAK)

    def init(self):
        """
            Init main application
        """
        self.settings = Settings.new()
        # Mount enclosing volume as soon as possible
        uris = self.settings.get_music_uris()
        try:
            for uri in uris:
                if uri.startswith("file:/"):
                    continue
                f = Gio.File.new_for_uri(uri)
                f.mount_enclosing_volume(Gio.MountMountFlags.NONE,
                                         None,
                                         None,
                                         None)
        except Exception as e:
            Logger.error("Application::init(): %s" % e)

        cssProviderFile = Gio.File.new_for_uri(
            "resource:///org/gnome/Lollypop/application.css")
        cssProvider = Gtk.CssProvider()
        cssProvider.load_from_file(cssProviderFile)
        screen = Gdk.Screen.get_default()
        styleContext = Gtk.StyleContext()
        styleContext.add_provider_for_screen(screen, cssProvider,
                                             Gtk.STYLE_PROVIDER_PRIORITY_USER)
        # Devel version, load devel theme
        if self.__app_id is None:
            css = "*:selected{ background-color: red; }"
            cssProvider = Gtk.CssProvider()
            cssProvider.load_from_data(css.encode("utf-8"))
            screen = Gdk.Screen.get_default()
            styleContext = Gtk.StyleContext()
            styleContext.add_provider_for_screen(
                screen, cssProvider, Gtk.STYLE_PROVIDER_PRIORITY_USER + 1)
        self.db = Database()
        self.cache = CacheDatabase()
        self.playlists = Playlists()
        self.albums = AlbumsDatabase(self.db)
        self.artists = ArtistsDatabase(self.db)
        self.genres = GenresDatabase(self.db)
        self.tracks = TracksDatabase(self.db)
        self.player = Player()
        self.inhibitor = Inhibitor()
        self.scanner = CollectionScanner()
        self.notify = NotificationManager()
        self.task_helper = TaskHelper()
        self.art_helper = ArtHelper()
        self.art = Artwork()
        self.art.update_art_size()
        self.album_art = AlbumArtwork()
        self.artist_art = ArtistArtwork()
        self.ws_director = DirectorWebService()
        self.ws_director.start()
        if not self.settings.get_value("disable-mpris"):
            from lollypop.mpris import MPRIS
            MPRIS(self)

        settings = Gtk.Settings.get_default()
        # Fallback setting
        dark = self.settings.get_value("dark-ui")
        if dark:
            settings.set_property(
                "gtk-application-prefer-dark-theme", dark)
        else:
            manager = Handy.StyleManager.get_default()
            if manager.get_system_supports_color_schemes():
                self.system_supports_color_schemes = True
                manager.set_color_scheme(Handy.ColorScheme.PREFER_LIGHT)
        ApplicationActions.__init__(self)
        monitor = Gio.NetworkMonitor.get_default()
        if monitor.get_network_available() and\
                not monitor.get_network_metered() and\
                self.settings.get_value("recent-youtube-dl"):
            self.task_helper.run(install_youtube_dl)

    def do_startup(self):
        """
            Init application
        """
        Gtk.Application.do_startup(self)
        Handy.init()
        if self.__window is None:
            from lollypop.window import Window
            self.init()
            self.__window = Window()
            self.__window.connect("delete-event", self.__hide_on_delete)
            self.__window.setup()
            self.__window.show()
            self.player.restore_state()

    def quit(self, vacuum=False, wait=100):
        """
            Quit Lollypop
            @param vacuum as bool
            @param wait as int
        """
        self.__window.container.stop()
        self.__window.hide()
        # Force stop after some tries
        if wait < 1000 and not self.ws_director.stop():
            GLib.timeout_add(wait, self.quit, vacuum, wait + 100)
            return
        self.album_art.cancellable.cancel()
        self.artist_art.cancellable.cancel()
        if self.settings.get_value("save-state"):
            self.__window.container.stack.save_history()
        # Then vacuum db
        if vacuum:
            self.__vacuum()
            self.art.clean_artwork()
        Gio.Application.quit(self)
        if GLib.environ_getenv(GLib.get_environ(), "DEBUG_LEAK") is not None:
            import gc
            gc.collect()
            for x in gc.garbage:
                s = str(x)
                print(type(x), "\n  ", s)

    def fullscreen(self):
        """
            Go fullscreen
        """
        def on_destroy(window):
            self.__fs_window = None
            self.__window.show()

        if self.__fs_window is None:
            from lollypop.fullscreen import FullScreen
            self.__fs_window = FullScreen()
            self.__fs_window.delayed_init()
            self.__fs_window.show()
            self.__fs_window.connect("destroy", on_destroy)
            self.__window.hide()
        else:
            self.__fs_window.destroy()

    @property
    def proxy_host(self):
        """
            Get proxy host
            @return str
        """
        return self.__proxy_host

    @property
    def proxy_port(self):
        """
            Get proxy port
            @return int
        """
        return self.__proxy_port

    @property
    def devices(self):
        """
            Get available devices
            Merge connected and known
            @return [str]
        """
        devices = self.__window.toolbar.end.devices_popover.devices
        devices += list(self.settings.get_value("devices"))
        result = []
        # Do not use set() + filter() because we want to keep order
        for device in devices:
            if device not in result and device != "":
                result.append(device)
        return result

    @property
    def is_fullscreen(self):
        """
            Return True if application is fullscreen
        """
        return self.__fs_window is not None

    @property
    def main_window(self):
        """
            Get main window
        """
        return self.__window

    @property
    def window(self):
        """
            Get current application window
            @return Gtk.Window
        """
        if self.__fs_window is not None:
            return self.__fs_window
        else:
            return self.__window

    @property
    def data_dir(self):
        """
            Get data dir
            @return str
        """
        return self.__data_dir

#######################
# PRIVATE             #
#######################
    def __save_state(self):
        """
            Save player state
        """
        if self.settings.get_value("save-state"):
            if self.player.current_track.id is None or\
                    self.player.current_track.storage_type &\
                    StorageType.EPHEMERAL:
                track_id = None
            else:
                track_id = self.player.current_track.id
                # Save albums context
                try:
                    with open(LOLLYPOP_DATA_PATH + "/Albums.bin", "wb") as f:
                        dump(self.player.albums, f)
                except Exception as e:
                    Logger.error("Application::__save_state(): %s" % e)
            dump(track_id, open(LOLLYPOP_DATA_PATH + "/track_id.bin", "wb"))
            dump([self.player.is_playing, self.player.is_party],
                 open(LOLLYPOP_DATA_PATH + "/player.bin", "wb"))
            dump(self.player.queue,
                 open(LOLLYPOP_DATA_PATH + "/queue.bin", "wb"))
            if self.player.current_track.id is not None:
                position = self.player.position
            else:
                position = 0
            dump(position, open(LOLLYPOP_DATA_PATH + "/position.bin", "wb"))
        self.player.stop_all()

    def __vacuum(self):
        """
            VACUUM DB
        """
        try:
            if self.scanner.is_locked():
                self.scanner.stop()
                GLib.idle_add(self.__vacuum)
                return
            SqlCursor.add(self.db)
            self.tracks.del_non_persistent(False)
            self.tracks.clean(False)
            self.albums.clean(False)
            self.artists.clean(False)
            self.genres.clean(False)
            SqlCursor.remove(self.db)
            self.cache.clean(True)

            with SqlCursor(self.db) as sql:
                sql.isolation_level = None
                sql.execute("VACUUM")
                sql.isolation_level = ""
            with SqlCursor(self.playlists) as sql:
                sql.isolation_level = None
                sql.execute("VACUUM")
                sql.isolation_level = ""
        except Exception as e:
            Logger.error("Application::__vacuum(): %s" % e)

    def __hide_on_delete(self, widget, event):
        """
            Hide window
            @param widget as Gtk.Widget
            @param event as Gdk.Event
        """
        # Quit if background mode is on but player is off
        if not self.settings.get_value("background-mode") or\
                not self.player.is_playing:
            GLib.idle_add(self.quit, True)
        return widget.hide_on_delete()

    def __on_activate(self, application):
        """
            Call default handler
            @param application as Gio.Application
        """
        self.__window.present()



# Copyright (c) 2018 Philipp Wolfer <ph.wolfer@gmail.com>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gi
from gi.repository import GObject

try:
    gi.require_version("Goa", "1.0")
    from gi.repository import Goa
except:
    pass

from lollypop.logger import Logger
from lollypop.utils import emit_signal


class GoaSyncedAccount(GObject.Object):
    """
        Provides access to a GOA account with given provider_name.

        The account will be kept in sync and will be updated if accounts
        are added, removed or changed.
    """

    __gsignals__ = {
        "account-switched": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self, provider_name):
        """
            Initialize a new GoaSyncedAccount for provider_name
            @param provider_name as str
        """
        GObject.Object.__init__(self)
        self._provider_name = provider_name
        self._proxy = None
        self._account = None
        self._oauth2_based = None
        try:
            self._client = Goa.Client.new_sync()
            self.__find_account()
            emit_signal(self, "account-switched")
            self._client.connect("account-added", self.__on_account_added)
            self._client.connect("account-removed", self.__on_account_removed)
            self._client.connect("account-changed", self.__on_account_changed)
        except:
            Logger.debug("GOA not available")
            self.__client = None

    @property
    def has_account(self):
        """
            True if there is an account with provider_name, False otherwise
            @return bool
        """
        return self._proxy is not None

    @property
    def account(self):
        """
            Return current GOA account
            @return Goa.Account
        """
        if self._proxy is None:
            return None
        if self._account is None:
            self._account = self._proxy.get_account()
        return self._account

    @property
    def oauth2_based(self):
        """
            Get OAuth2Based for current account
            @return Goa.OAuth2Based
        """
        if self._proxy is None:
            return None
        if self._oauth2_based is None:
            self._oauth2_based = self._proxy.get_oauth2_based()
        return self._oauth2_based

#######################
# PRIVATE             #
#######################
    def __find_account(self):
        """
            Find account matching current provider
        """
        Logger.debug("GOA __find_account")
        self._proxy = None
        try:
            for proxy in self._client.get_accounts():
                if self.__account_matches_provider(proxy):
                    Logger.debug("GOA account found")
                    self._proxy = proxy
                    return
        except Exception as e:
            Logger.debug("GOA __find_account failed: %s" % e)
            pass

    def __account_matches_provider(self, proxy):
        """
            True if current account match proxy account provider
            @param proxy as Goa.Object
            @return bool
        """
        account = proxy.get_account()
        Logger.debug("GOA __account_matches_provider: %s = %s ?" %
                     (account.props.provider_name, self._provider_name))
        return account.props.provider_name == self._provider_name

    def __on_account_added(self, client, proxy):
        """
            Update proxy and emit account-switched
            @param client as Goa.Client
            @param proxy as Goa.Object
        """
        Logger.debug("GOA account added")
        if self._proxy is None and self.__account_matches_provider(proxy):
            self._proxy = proxy
            emit_signal(self, "account-switched")

    def __on_account_removed(self, client, proxy):
        """
            Try finding a new account and emit account-switched
            @param client as Goa.Client
            @param proxy as Goa.Object
        """
        Logger.debug("GOA account removed")
        if self._proxy == proxy:
            self.__find_account()
            emit_signal(self, "account-switched")

    def __on_account_changed(self, client, proxy):
        """
            Reset current account settings
            @param client as Goa.Client
            @param proxy as Goa.Object
        """
        Logger.debug("GOA account changed")
        if self._proxy == proxy:
            self._account = None
            self._oauth2_based = None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from locale import strcoll

from lollypop.view_lazyloading import LazyLoadingView
from lollypop.helper_gestures import GesturesHelper
from lollypop.define import ViewType, App
from lollypop.utils import get_font_height, popup_widget, set_cursor_type


class FlowBoxView(LazyLoadingView, GesturesHelper):
    """
        Lazy loading FlowBox
    """

    def __init__(self, storage_type, view_type=ViewType.SCROLLED):
        """
            Init flowbox view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        LazyLoadingView.__init__(self, storage_type, view_type)
        self._items = []
        self.__hovered_child = None
        self.__font_height = get_font_height()
        self._box = Gtk.FlowBox()
        self._box.get_style_context().add_class("small_padding")
        # Allow lazy loading to not jump up and down
        self._box.set_homogeneous(True)
        self._box.set_vexpand(True)
        self._box.set_max_children_per_line(1000)
        self._box.set_property("valign", Gtk.Align.START)
        self._box.connect("child-activated", self._on_child_activated)
        self._box.show()
        if App().animations:
            self.__event_controller = Gtk.EventControllerMotion.new(self._box)
            self.__event_controller.connect("motion", self.__on_box_motion)
        GesturesHelper.__init__(self, self._box)

    def populate(self, items):
        """
            Populate items
            @param items
        """
        LazyLoadingView.populate(self, items)

    def add_value(self, value):
        """
            Append value
            @param value as object
        """
        self._box.set_sort_func(self._sort_func)
        self.add_value_unsorted(value)

    def add_value_unsorted(self, value):
        """
            Add value unsorted
            @param value as object
        """
        child = self._get_child(value)
        child.populate()

    def remove_value(self, value):
        """
            Remove value
            @param value as object
        """
        for child in self._box.get_children():
            if child.data == value:
                child.destroy()
                break

    def clear(self):
        """
            Clear flowbox
        """
        for child in self._box.get_children():
            child.destroy()

    def destroy(self):
        """
            Force destroying the box
            Help freeing memory, no idea why
        """
        self._box.destroy()
        LazyLoadingView.destroy(self)

    @property
    def font_height(self):
        """
            Get font height
            @return int
        """
        return self.__font_height

    @property
    def children(self):
        """
            Get box children
            @return [Gtk.Widget]
        """
        return self._box.get_children()

#######################
# PROTECTED           #
#######################
    def _get_menu_widget(self, child):
        """
            Get menu widget
            @param child as Gtk.FlowBoxChild
            @return Gtk.Widget
        """
        return None

    def _get_label_height(self):
        """
            Get wanted label height
            @return int
        """
        return 0

    def _sort_func(self, widget1, widget2):
        """
            Sort function
            @param widget1 as Gtk.Widget
            @param widget2 as Gtk.Widget
        """
        return strcoll(widget1.sortname, widget2.sortname)

    def _on_child_activated(self, flowbox, child):
        pass

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        LazyLoadingView._on_container_folded(self, leaflet, folded)
        self.pause()
        children = self._box.get_children()
        for child in children:
            child.reset_artwork()
            self.queue_lazy_loading(child)
        self.lazy_loading()

    def _on_view_leave(self, event_controller):
        """
            Unselect selected child
            @param event_controller as Gtk.EventControllerMotion
        """
        self.__unselect_selected()

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Popup menu at position
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        self._on_primary_long_press_gesture(x, y)

    def _on_primary_long_press_gesture(self, x, y):
        """
            Popup menu at position
            @param x as int
            @param y as int
        """
        child = self._box.get_child_at_pos(x, y)
        if child is None or child.artwork is None:
            return
        self.__popup_menu(child)

    def _on_destroy(self, widget):
        """
            Clean up widget
            @param widget as Gtk.Widget
        """
        LazyLoadingView._on_destroy(self, widget)
        self.__event_controller = None

#######################
# PRIVATE             #
#######################
    def __popup_menu(self, child):
        """
            Popup album menu at position
            @param child ad RoundedArtistWidget
        """
        menu_widget = self._get_menu_widget(child)
        if menu_widget is not None:
            menu_widget.show()
            popup_widget(menu_widget, child.artwork, None, None, None)

    def __unselect_selected(self):
        """
            Unselect selected child
        """
        if self.__hovered_child is not None and\
                self.__hovered_child.artwork is not None:
            self.__hovered_child.artwork.unset_state_flags(
                Gtk.StateFlags.VISITED)
            set_cursor_type(self.__hovered_child, "left_ptr")
            self.__hovered_child = None

    def __on_box_motion(self, event_controller, x, y):
        """
            Update current selected child
            @param event_controller as Gtk.EventControllerMotion
            @param x as int
            @param y as int
        """
        child = self._box.get_child_at_pos(x, y)
        if child == self.__hovered_child:
            return
        elif child is not None and child.artwork is not None:
            child.artwork.set_state_flags(Gtk.StateFlags.VISITED, False)
            self.__unselect_selected()
            self.__hovered_child = child
            set_cursor_type(child)
        else:
            self.__unselect_selected()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk

from lollypop.define import ViewType, App
from lollypop.utils import noaccents
from lollypop.utils_album import tracks_to_albums
from lollypop.logger import Logger


class FilteringHelper():
    """
        Helper for filtering widgets Boxes
    """

    def __init__(self):
        """
            Init helper
        """
        self.__last_scrolled = None

    def search_for_child(self, text):
        """
            Search child and scroll
            @param text as str
        """
        for child in self.filtered:
            style_context = child.get_style_context()
            style_context.remove_class("typeahead")
        if not text:
            return
        for child in self.filtered:
            if noaccents(child.name).find(noaccents(text)) != -1:
                style_context = child.get_style_context()
                style_context.add_class("typeahead")
                GLib.idle_add(self._scroll_to_child, child)
                break

    def search_prev(self, text):
        """
            Search previous child and scroll
            @param text as str
        """
        previous_children = []
        found_child = None
        for child in self.filtered:
            style_context = child.get_style_context()
            if style_context.has_class("typeahead"):
                found_child = child
                break
            previous_children.insert(0, child)
        if previous_children and found_child is not None:
            for child in previous_children:
                if noaccents(child.name).find(noaccents(text)) != -1:
                    found_child.get_style_context().remove_class("typeahead")
                    child.get_style_context().add_class("typeahead")
                    GLib.idle_add(self._scroll_to_child, child)
                    break

    def search_next(self, text):
        """
            Search previous child and scroll
            @param text as str
        """
        found = False
        previous_style_context = None
        for child in self.filtered:
            style_context = child.get_style_context()
            if style_context.has_class("typeahead"):
                previous_style_context = style_context
                found = True
                continue
            if found and noaccents(child.name).find(noaccents(text)) != -1:
                previous_style_context.remove_class("typeahead")
                style_context.add_class("typeahead")
                GLib.idle_add(self._scroll_to_child, child)
                break

    def activate_child(self):
        """
            Activated typeahead row
        """
        def reset_party_mode():
            if App().player.is_party:
                App().lookup_action("party").change_state(
                    GLib.Variant("b", False))

        try:
            # Search typeahead child
            typeahead_child = None
            for child in self.filtered:
                style_context = child.get_style_context()
                if style_context.has_class("typeahead"):
                    typeahead_child = child
                    break
            if typeahead_child is None:
                return
            from lollypop.view_current_albums import CurrentAlbumsView
            from lollypop.view_playlists import PlaylistsView
            # Play child without reseting player
            if isinstance(self, CurrentAlbumsView):
                if hasattr(typeahead_child, "album"):
                    album = typeahead_child.album
                    if album.tracks:
                        track = album.tracks[0]
                        App().player.load(track)
                elif hasattr(typeahead_child, "track"):
                    App().player.load(typeahead_child.track)
            # Play playlist
            elif isinstance(self, PlaylistsView):
                reset_party_mode()
                tracks = []
                for album_row in self.view.children:
                    for track in album_row.album.tracks:
                        tracks.append(track)
                if tracks:
                    albums = tracks_to_albums(tracks)
                    if hasattr(typeahead_child, "album"):
                        if typeahead_child.album.tracks:
                            App().player.play_track_for_albums(
                                typeahead_child.album.tracks[0], albums)
                    elif hasattr(typeahead_child, "track"):
                        App().player.play_track_for_albums(
                            typeahead_child.track, albums)
            elif hasattr(typeahead_child, "data"):
                reset_party_mode()
                typeahead_child.activate()
            elif hasattr(typeahead_child, "album"):
                reset_party_mode()
                album = typeahead_child.album
                App().player.add_album(album)
                App().player.load(album.tracks[0])
            elif hasattr(typeahead_child, "track"):
                reset_party_mode()
                track = typeahead_child.track
                App().player.add_album(track.album)
                App().player.load(track.album.get_track(track.id))
            App().window.container.type_ahead.entry.set_text("")
        except Exception as e:
            Logger.error("View::activate_child: %s" % e)

    @property
    def filtered(self):
        """
            Return widget that we should filter
            @return [Gtk.Widget]
        """
        return self.children

    @property
    def scroll_shift(self):
        """
            Get scroll shift for y axes
            @return int
        """
        return 0

#######################
# PROTECTED           #
#######################
    def _scroll_to_child(self, child):
        """
            Scroll to child
            @param child as Gtk.Widget
        """
        if self.view_type & ViewType.SCROLLED:
            view_widget = self.scrolled.get_child()
            # Filtered children may exists while view is not populated
            if view_widget is None:
                return
            if isinstance(view_widget, Gtk.Viewport):
                view_widget = view_widget.get_child()
            coordinates = child.translate_coordinates(
                view_widget, 0, -self.scroll_shift)
            if coordinates:
                self.scrolled.get_vadjustment().set_value(coordinates[1])

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib


from lollypop.utils import set_cursor_type, popup_widget
from lollypop.widgets_player_artwork import ArtworkPlayerWidget
from lollypop.widgets_player_label import LabelPlayerWidget
from lollypop.define import App, ArtBehaviour, StorageType, MARGIN_SMALL
from lollypop.define import ViewType
from lollypop.helper_gestures import GesturesHelper


class ToolbarInfo(Gtk.Bin, ArtworkPlayerWidget, GesturesHelper):
    """
        Informations toolbar
    """

    def __init__(self):
        """
            Init toolbar
        """
        Gtk.Bin.__init__(self)
        self.__width = 0
        horizontal_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 15)
        horizontal_box.show()
        self.__eventbox = Gtk.EventBox.new()
        self.__eventbox.add(horizontal_box)
        self.__eventbox.set_property("halign", Gtk.Align.START)
        self.__eventbox.show()
        self.__eventbox.connect("realize", set_cursor_type)
        self.add(self.__eventbox)
        GesturesHelper.__init__(self, self.__eventbox)
        self.special_headerbar_hack()

        self.__label = LabelPlayerWidget()
        self.__artwork = ArtworkPlayerWidget(ArtBehaviour.CROP_SQUARE |
                                             ArtBehaviour.CACHE)
        self.__artwork.set_margin_top(1)
        horizontal_box.pack_start(self.__artwork, False, False, 0)
        horizontal_box.pack_start(self.__label, False, False, 0)
        self.set_margin_start(MARGIN_SMALL)
        self.connect("realize", self.__on_realize)
        self.set_property("has-tooltip", True)
        self.connect("query-tooltip", self.__on_query_tooltip)

    def show_children(self):
        """
            Show labels and artwork
        """
        self.__artwork.show()
        self.__label.show()

    def hide_children(self):
        """
            Hide labels and artwork, ignore self
        """
        self.__artwork.hide()
        self.__label.hide()

    def do_get_preferred_width(self):
        """
            We force preferred width
            @return (int, int)
        """
        return (self.__width, self.__width)

    def get_preferred_height(self):
        """
            Return preferred height
            @return (int, int)
        """
        return self.__labels.get_preferred_height()

    def set_width(self, width):
        """
            Set widget width
            @param width as int
        """
        self.__width = width
        self.set_property("width-request", width)

#######################
# PROTECTED           #
#######################
    def _on_primary_long_press_gesture(self, x, y):
        """
            Show menu
            @param x as int
            @param y as int
        """
        if App().window.folded or not self.__artwork.get_visible():
            return
        if App().player.current_track.id is not None:
            self.__popup_menu()

    def _on_primary_press_gesture(self, x, y, event):
        """
            Show information popover
            @param x as int
            @param y as int
            @param evnet as Gdk.Event
        """
        if App().window.folded or not self.__artwork.get_visible():
            return
        if App().player.current_track.id is not None:
            from lollypop.pop_information import InformationPopover
            popover = InformationPopover()
            popover.populate()
        popover.set_relative_to(self.__eventbox)
        popover.popup()

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Show menu
            @param x as int
            @param y as int
        """
        self._on_primary_long_press_gesture(x, y)

#######################
# PRIVATE             #
#######################
    def __popup_menu(self):
        """
            Show contextual menu
        """
        if App().window.folded or not self.__artwork.get_visible():
            return
        track = App().player.current_track
        if track.id is not None and track.id >= 0:
            from lollypop.menu_objects import TrackMenu, TrackMenuExt
            from lollypop.widgets_menu import MenuBuilder
            menu = TrackMenu(track, ViewType.TOOLBAR)
            menu_widget = MenuBuilder(menu)
            menu_widget.show()
            if not track.storage_type & StorageType.EPHEMERAL:
                menu_ext = TrackMenuExt(track)
                menu_ext.show()
                menu_widget.add_widget(menu_ext)
            self.set_state_flags(Gtk.StateFlags.FOCUSED, False)
            popup_widget(menu_widget, self.__eventbox, None, None, None)

    def __on_query_tooltip(self, widget, x, y, keyboard, tooltip):
        """
            Show tooltip if needed
            @param widget as Gtk.Widget
            @param x as int
            @param y as int
            @param keyboard as bool
            @param tooltip as Gtk.Tooltip
        """
        track = App().player.current_track
        if track.id is not None and track.id >= 0:
            artist = GLib.markup_escape_text(",".join(track.artists))
            title = GLib.markup_escape_text(track.title)
            album = GLib.markup_escape_text(track.album.name)
            tooltip.set_markup("<b>%s</b> - %s\n%s" % (title, artist, album))
            return True
        return False

    def __on_realize(self, toolbar):
        """
            Calculate art size
            @param toolbar as ToolbarInfos
        """
        art_size = self.get_allocated_height() - 1
        self.__artwork.set_art_size(art_size, art_size)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from lollypop.define import App, Type


def tracks_to_albums(tracks, skipped=True):
    """
        Convert tracks list to albums list
        @param tracks as [Track]
        @param skipped as bool
        @return [Album]
    """
    albums = []
    for track in tracks:
        if albums and albums[-1].id == track.album.id:
            albums[-1].append_track(track, False)
        else:
            album = track.album
            if album.loved == -1 and not skipped:
                continue
            album.set_tracks([track], False)
            albums.append(album)
    return albums


def get_album_ids_for(genre_ids, artist_ids, storage_type, skipped):
    """
        Get album ids view for genres/artists
        @param genre_ids as [int]
        @param artist_ids as [int]
        @param storage_type as StorageType
        @return [int]
    """
    items = []
    limit = App().settings.get_value("view-limit").get_int32()
    if genre_ids and genre_ids[0] == Type.POPULARS:
        items = App().albums.get_rated(storage_type, skipped, limit)
        count = limit - len(items)
        for album in App().albums.get_populars(storage_type, skipped, count):
            if album not in items:
                items.append(album)
    elif genre_ids and genre_ids[0] == Type.LOVED:
        items = App().albums.get_loved_albums(storage_type)
    elif genre_ids and genre_ids[0] == Type.RECENTS:
        items = App().albums.get_recents(storage_type, skipped, limit)
    elif genre_ids and genre_ids[0] == Type.LITTLE:
        items = App().albums.get_little_played(storage_type, skipped, limit)
    elif genre_ids and genre_ids[0] == Type.RANDOMS:
        items = App().albums.get_randoms(storage_type, None, skipped, limit)
    elif genre_ids and genre_ids[0] == Type.COMPILATIONS:
        items = App().albums.get_compilation_ids([], storage_type, skipped)
    elif genre_ids and not artist_ids:
        if Type.WEB in genre_ids or\
                App().settings.get_value("show-compilations-in-album-view"):
            items = App().albums.get_compilation_ids(genre_ids, storage_type,
                                                     skipped)
        items += App().albums.get_ids(genre_ids, [], storage_type, skipped)
    else:
        items = App().albums.get_ids(genre_ids, artist_ids,
                                     storage_type, skipped)
    return items



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from lollypop.define import Type, App


class ShownLists:
    """
        Handle shown lists
    """

    IDS = {
        Type.SUGGESTIONS: _("Suggestions"),
        Type.POPULARS: _("Popular albums"),
        Type.RANDOMS: _("Random albums"),
        Type.LOVED: _("Loved albums"),
        Type.RECENTS: _("Recently added albums"),
        Type.LITTLE: _("Seldomly played albums"),
        Type.PLAYLISTS: _("Playlists"),
        Type.YEARS: _("Years"),
        Type.GENRES: _("Genres"),
        Type.GENRES_LIST: _("Genres (list)"),
        Type.WEB: _("Web"),
        Type.LYRICS: _("Lyrics"),
        Type.ALL: _("Albums"),
        Type.ARTISTS: _("Artists"),
        Type.ARTISTS_LIST: _("Artists (list)"),
        Type.SEARCH: _("Search"),
        Type.CURRENT: _("Playing albums"),
        Type.INFO: _("Information"),
        Type.COMPILATIONS: _("Compilations"),
    }

    def get(mask, get_all=False):
        """
            Get list
            @param mask as bit mask
            @param get_all as bool
            @return [(,)]
        """
        wanted = list(App().settings.get_value("shown-album-lists"))
        lists = []
        for key in ShownLists.IDS.keys():
            string = ShownLists.IDS[key]
            if get_all or key in wanted:
                lists.append((key, string, ""))
        lists.append((Type.SEPARATOR, "", ""))
        lists.sort(key=lambda tup: tup[0], reverse=True)
        return lists


class ShownPlaylists(ShownLists):
    """
        Handle shown playlists
    """
    IDS = {
        Type.POPULARS: _("Popular tracks"),
        Type.RANDOMS: _("Random tracks"),
        Type.LOVED: _("Loved tracks"),
        Type.RECENTS: _("Recently played tracks"),
        Type.LITTLE: _("Seldomly played tracks"),
        Type.SKIPPED: _("Skipped tracks"),
        Type.ALL: _("All tracks")
    }

    def get(get_all=False):
        """
            Get list
            @return [(,)]
        """
        wanted = App().settings.get_value("shown-playlists")
        lists = []
        for key in ShownPlaylists.IDS.keys():
            string = ShownPlaylists.IDS[key]
            if get_all or key in wanted:
                lists.append((key, string, ""))
        lists.sort(key=lambda tup: tup[0], reverse=True)
        return lists



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Gio, Pango

from gettext import gettext as _

from lollypop.define import App, Type, MARGIN, ViewType, StorageType
from lollypop.objects_album import Album
from lollypop.utils import get_network_available, get_default_storage_type
from lollypop.helper_signals import signals
from lollypop.helper_horizontal_scrolling import HorizontalScrollingHelper
from lollypop.view_albums_box import AlbumsBoxView


class AlbumsLineView(AlbumsBoxView, HorizontalScrollingHelper):
    """
        Albums on a line
    """

    ITEMS = 20

    def __init__(self, storage_type, view_type):
        """
            Init view
            @param view_type as ViewType
            @param storage_type as StorageType
        """
        AlbumsBoxView.__init__(self, [], [], storage_type, view_type)
        self.set_property("valign", Gtk.Align.START)
        self._label = Gtk.Label.new()
        self._label.set_ellipsize(Pango.EllipsizeMode.END)
        self._label.set_hexpand(True)
        self._label.set_property("halign", Gtk.Align.START)
        self._label.get_style_context().add_class("dim-label")
        self.__update_label(App().window.folded)
        self._backward_button = Gtk.Button.new_from_icon_name(
                                                        "go-previous-symbolic",
                                                        Gtk.IconSize.BUTTON)
        self._forward_button = Gtk.Button.new_from_icon_name(
                                                       "go-next-symbolic",
                                                       Gtk.IconSize.BUTTON)
        self._backward_button.get_style_context().add_class("menu-button")
        self._forward_button.get_style_context().add_class("menu-button")
        header = Gtk.Grid()
        header.set_column_spacing(10)
        header.add(self._label)
        header.add(self._backward_button)
        header.add(self._forward_button)
        header.set_margin_end(MARGIN)
        header.show_all()
        HorizontalScrollingHelper.__init__(self)
        self.add(header)
        self.add_widget(self._box)

    def populate(self, albums):
        """
            Configure widget based on albums
            @param items as [Album]
        """
        if albums:
            self.show()
            self._box.set_min_children_per_line(len(albums))
            AlbumsBoxView.populate(self, albums)

    def add_value(self, album):
        """
            Add a new album
            @param album as Album
        """
        AlbumsBoxView.add_value_unsorted(self, album)
        self._box.set_min_children_per_line(len(self._box.get_children()))
        self.update_buttons()

    @property
    def args(self):
        return None

#######################
# PROTECTED           #
#######################
    def _on_collection_updated(self, scanner, item, scan_update):
        pass

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        AlbumsBoxView._on_container_folded(self, leaflet, folded)
        self.__update_label(folded)

    def _on_populated(self, widget):
        """
            Update buttons
            @param widget as Gtk.Widget
        """
        self.update_buttons()
        AlbumsBoxView._on_populated(self, widget)

#######################
# PRIVATE             #
#######################
    def __update_label(self, folded):
        """
            Update label style based on current adaptive state
            @param folded as bool
        """
        style_context = self._label.get_style_context()
        if folded:
            style_context.remove_class("text-x-large")
        else:
            style_context.add_class("text-x-large")


class AlbumsArtistLineView(AlbumsLineView):
    """
        Artist album line
    """

    def __init__(self,  artist_id, genre_ids, storage_type, view_type):
        """
            Init view
            @param artist_id as artist_id
            @param genre_ids as [int]
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsLineView.__init__(self, storage_type, view_type)
        self.__artist_id = artist_id
        self.__genre_ids = genre_ids

    def populate(self, excluded_album_id):
        """
            Populate view less excluded_album_id
            @param excluded_album_id as int
        """
        def on_load(items):
            AlbumsLineView.populate(self, items)

        def load():
            if self.__artist_id == Type.COMPILATIONS:
                album_ids = App().albums.get_compilation_ids(
                    self.__genre_ids, self.storage_type, True)
            else:
                album_ids = App().albums.get_ids(
                    self.__genre_ids, [self.__artist_id],
                    self.storage_type, True)
            if excluded_album_id in album_ids:
                album_ids.remove(excluded_album_id)
            return [Album(album_id) for album_id in album_ids]

        if self.__artist_id == Type.COMPILATIONS:
            self._label.set_text(_("Others compilations"))
        else:
            self._label.set_text(App().artists.get_name(self.__artist_id))
        App().task_helper.run(load, callback=(on_load,))


class AlbumsArtistAppearsOnLineView(AlbumsLineView):
    """
        Show albums where artist is in featured
    """

    def __init__(self,  artist_ids, genre_ids, storage_type, view_type):
        """
            Init view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsLineView.__init__(self, storage_type, view_type)
        self.__artist_ids = artist_ids
        self.__genre_ids = genre_ids

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            AlbumsLineView.populate(self, items)

        def load():
            album_ids = App().artists.get_featured(self.__genre_ids,
                                                   self.__artist_ids,
                                                   self.storage_type,
                                                   True)
            return [Album(album_id) for album_id in album_ids]

        self._label.set_text(_("Appears on"))
        App().task_helper.run(load, callback=(on_load,))


class AlbumsPopularsLineView(AlbumsLineView):
    """
        Populars albums line
    """

    def __init__(self, storage_type, view_type):
        """
            Init view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsLineView.__init__(self, storage_type, view_type)

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            AlbumsLineView.populate(self, items)

        def load():
            storage_type = get_default_storage_type()
            album_ids = App().albums.get_populars_at_the_moment(storage_type,
                                                                False,
                                                                self.ITEMS)
            return [Album(album_id) for album_id in album_ids]

        self._label.set_text(_("Popular albums at the moment"))
        App().task_helper.run(load, callback=(on_load,))


class AlbumsRandomGenresLineView(AlbumsLineView):
    """
        Populars albums line
    """

    def __init__(self, storage_type, view_type):
        """
            Init view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsLineView.__init__(self, storage_type, view_type)

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            AlbumsLineView.populate(self, items)

        def load():
            (genre_id, genre) = App().genres.get_random()
            GLib.idle_add(self._label.set_text, genre)
            storage_type = get_default_storage_type()
            album_ids = App().albums.get_randoms(storage_type,
                                                 genre_id,
                                                 False,
                                                 self.ITEMS)
            return [Album(album_id) for album_id in album_ids]

        App().task_helper.run(load, callback=(on_load,))


class AlbumsSearchLineView(AlbumsLineView):
    """
        Search album line
    """
    def __init__(self, storage_type):
        """
            Init view
            @param storage_type as StorageType
        """
        AlbumsLineView.__init__(self, storage_type, ViewType.SEARCH |
                                ViewType.SCROLLED | ViewType.ALBUM)
        self.__album_ids = []
        self._label.set_text(_("Albums"))

    def add_value(self, album):
        """
            Add a new album
            @param album as Album
        """
        if album.id in self.__album_ids:
            return
        self.__album_ids.append(album.id)
        AlbumsLineView.add_value(self, album)

    def clear(self):
        """
            Clear and hide the view
        """
        self.__album_ids = []
        AlbumsLineView.clear(self)
        GLib.idle_add(self.hide)


class AlbumsStorageTypeLineView(AlbumsLineView):
    """
        Storage type album line
    """

    @signals
    def __init__(self, text, storage_type, view_type):
        """
            Init view
            @param text as str
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsLineView.__init__(self, storage_type, view_type)
        self._label.set_text(text)
        self.__cancellable = Gio.Cancellable()
        self.__storage_type = StorageType.NONE
        return [
                (App().ws_director.collection_ws, "match-album",
                 "_on_album_match"),
                (App().settings, "changed::network-access",
                 "_on_network_access_changed"),
                (App().settings, "changed::network-access-acl",
                 "_on_network_access_changed")
        ]

    def populate(self, storage_type):
        """
            Populate view
            @param storage_type as StorageType
        """
        def on_load(items):
            AlbumsLineView.populate(self, items)

        def load():
            album_ids = App().albums.get_for_storage_type(storage_type, 20)
            return [Album(album_id) for album_id in album_ids]

        App().task_helper.run(load, callback=(on_load,))
        self.__storage_type |= storage_type

#######################
# PROTECTED           #
#######################
    def _on_unmap(self, widget):
        """
            Cancel search
            @param widget as Gtk.Widget
        """
        self.__cancellable.cancel()
        self.__cancellable = Gio.Cancellable()

    def _on_album_match(self, spotify_helper, album_id, storage_type):
        """
            Handles changes in collection
            @param scanner as CollectionScanner
            @param album_id as int
            @param storage_type as StorageType
        """
        count = len(self.children)
        if count == self.ITEMS:
            return
        if self.__storage_type & storage_type:
            self.add_value(Album(album_id))
            self.show()

    def _on_network_access_changed(self, *ignore):
        """
            Destroy if not allowed anymore
        """
        if not get_network_available("SPOTIFY") or\
                not get_network_available("YOUTUBE"):
            self.destroy()

#####################
# PRIVATE           #
#####################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, Gtk

from lollypop.define import App, ScanType, Type, LovedFlags


class ApplicationActions:
    """
        Application actions
    """

    def __init__(self):
        """
            Init actions
        """
        self.__restore_mini_player = False
        self.__special_shortcuts_count = 0
        settings_action = Gio.SimpleAction.new("settings", None)
        settings_action.connect("activate", self.__on_settings_activate)
        App().add_action(settings_action)

        update_action = Gio.SimpleAction.new("update_db", None)
        update_action.connect("activate", self.__on_update_db_activate)
        App().add_action(update_action)

        fullscreen_action = Gio.SimpleAction.new("fullscreen", None)
        fullscreen_action = Gio.SimpleAction.new_stateful(
                    "fullscreen",
                    None,
                    GLib.Variant.new_boolean(False))
        App().player.connect("status-changed",
                             self.__on_status_changed,
                             fullscreen_action)
        fullscreen_action.set_enabled(False)
        fullscreen_action.connect("change-state",
                                  self.__on_fullscreen_change_state)
        App().add_action(fullscreen_action)

        equalizer_action = Gio.SimpleAction.new("equalizer", None)
        equalizer_action.connect("activate", self.__on_equalizer_activate)
        App().set_accels_for_action("app.equalizer", ["<Shift><Alt>e"])
        App().add_action(equalizer_action)

        miniplayer_action = Gio.SimpleAction.new_stateful(
                    "miniplayer",
                    None,
                    GLib.Variant.new_boolean(False))
        App().player.connect("status-changed",
                             self.__on_status_changed,
                             miniplayer_action)
        miniplayer_action.set_enabled(False)
        miniplayer_action.connect("change-state",
                                  self.__on_miniplayer_change_state)
        App().set_accels_for_action("app.miniplayer", ["<Shift><Alt>m"])
        App().add_action(miniplayer_action)

        about_action = Gio.SimpleAction.new("about", None)
        about_action.connect("activate", self.__on_about_activate)
        App().add_action(about_action)

        shortcuts_action = Gio.SimpleAction.new("shortcuts", None)
        shortcuts_action.connect("activate", self.__on_shortcuts_activate)
        App().add_action(shortcuts_action)

        quit_action = Gio.SimpleAction.new("quit", None)
        quit_action.connect("activate", lambda x, y: App().quit(True))
        App().add_action(quit_action)

        seek_action = Gio.SimpleAction.new("seek",
                                           GLib.VariantType.new("i"))
        seek_action.connect("activate", self.__on_seek_action)
        App().add_action(seek_action)
        player_action = Gio.SimpleAction.new("shortcut",
                                             GLib.VariantType.new("s"))
        player_action.connect("activate", self.__on_player_action)
        App().add_action(player_action)

        search_action = Gio.SimpleAction.new("search",
                                             GLib.VariantType.new("s"))
        App().add_action(search_action)
        App().set_accels_for_action("app.search('')", ["<Control>f"])
        search_action.connect("activate", self.__on_search_activate)

        # Special action to queue a view reload
        reload_action = Gio.SimpleAction.new_stateful(
                "reload",
                None,
                GLib.Variant.new_boolean(False))
        App().add_action(reload_action)

        self.__setup_global_shortcuts()
        self.enable_special_shortcuts(True)

    def enable_special_shortcuts(self, enable):
        """
            Enable special shortcuts
            @param enable as bool
        """
        if enable:
            if self.__special_shortcuts_count == 0:
                if Gtk.Widget.get_default_direction() == Gtk.TextDirection.RTL:
                    App().set_accels_for_action("app.seek(10000)", ["Left"])
                    App().set_accels_for_action("app.seek(-10000)", ["Right"])
                else:
                    App().set_accels_for_action("app.seek(10000)", ["Right"])
                    App().set_accels_for_action("app.seek(-10000)", ["Left"])
                App().set_accels_for_action("app.shortcut::play_pause",
                                            ["c", "space"])
                App().set_accels_for_action("app.shortcut::play", ["x"])
                App().set_accels_for_action("app.shortcut::stop", ["v"])
                App().set_accels_for_action("app.shortcut::next", ["n"])
                App().set_accels_for_action("app.shortcut::prev", ["p"])
            self.__special_shortcuts_count += 1
        else:
            self.__special_shortcuts_count -= 1
            if self.__special_shortcuts_count == 0:
                App().set_accels_for_action("app.seek(10000)", [])
                App().set_accels_for_action("app.seek(-10000)", [])
                App().set_accels_for_action("app.shortcut::play_pause", [])
                App().set_accels_for_action("app.shortcut::play", [])
                App().set_accels_for_action("app.shortcut::stop", [])
                App().set_accels_for_action("app.shortcut::next", [])
                App().set_accels_for_action("app.shortcut::prev", [])

#######################
# PRIVATE             #
#######################
    def __setup_global_shortcuts(self):
        """
            Setup global shortcuts
        """
        if Gtk.Widget.get_default_direction() == Gtk.TextDirection.RTL:
            App().set_accels_for_action("app.shortcut::go_back",
                                        ["<Alt>Right", "Back"])
        else:
            App().set_accels_for_action("app.shortcut::go_back",
                                        ["<Alt>Left", "Back"])
        App().set_accels_for_action("app.shortcut::filter", ["<Control>i"])
        App().set_accels_for_action("app.shortcut::volume",
                                    ["<Shift><Alt>v"])
        App().set_accels_for_action("app.shortcut::lyrics",
                                    ["<Shift><Alt>l"])
        App().set_accels_for_action("app.shortcut::next_album", ["<Control>n"])
        App().set_accels_for_action("app.update_db", ["<Control>u"])
        App().set_accels_for_action("app.settings", ["<Control>comma"])
        App().set_accels_for_action("app.fullscreen", ["F11", "F7"])
        App().set_accels_for_action("app.mini", ["<Control>m"])
        App().set_accels_for_action("app.about", ["F3"])
        App().set_accels_for_action("app.shortcuts", ["F2"])
        App().set_accels_for_action("app.help", ["F1"])
        App().set_accels_for_action("app.quit", ["<Control>q"])
        App().set_accels_for_action("app.shortcut::loved", ["<Alt>l"])
        App().set_accels_for_action("app.shortcut::reload", ["<Control>r"])
        App().set_accels_for_action("app.shortcut::volume_up",
                                    ["<Shift><Alt>Up"])
        App().set_accels_for_action("app.shortcut::volume_down",
                                    ["<Shift><Alt>Down"])

    def __on_update_db_activate(self, *ignore):
        """
            Search for new music
        """
        if App().window:
            App().scanner.update(ScanType.FULL)

    def __on_about_activate_response(self, dialog, response_id):
        """
            Destroy about dialog when closed
            @param dialog as Gtk.Dialog
            @param response id as int
        """
        dialog.destroy()

    def __on_search_activate(self, action, value):
        """
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        search = value.get_string()
        App().window.container.show_view([Type.SEARCH], search)

    def __on_fullscreen_change_state(self, action, value):
        """
            Show a fullscreen window with cover and artist information
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        action.set_state(value)
        App().fullscreen()

    def __on_equalizer_activate(self, action, value):
        """
            Show equalizer view
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        App().window.container.show_view([Type.EQUALIZER])

    def __on_miniplayer_change_state(self, action, value):
        """
            Show miniplayer view
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        def replace_window():
            App().window.hide()
            App().window.show()
            App().window.set_opacity(1)

        App().window.set_opacity(0)
        action.set_state(value)
        if value:
            self.__restore_mini_player = App().window.miniplayer is not None
            App().window.show_miniplayer(True, True)
            App().window.toolbar.hide_info_and_buttons(True)
            App().window.unmaximize()
            App().window.resize(1, 1)
        else:
            if self.__restore_mini_player:
                App().window.miniplayer.reveal(False)
            else:
                App().window.miniplayer.reveal(False)
                App().window.show_miniplayer(False)
                App().window.toolbar.hide_info_and_buttons(False)
            size = App().settings.get_value("window-size")
            maximized = App().settings.get_value("window-maximized")
            App().window.resize(size[0], size[1])
            if maximized:
                GLib.idle_add(App().window.maximize)
        # Big hack, wait unmaximize/maximize to finish
        GLib.timeout_add(250, replace_window)

    def __on_settings_activate(self, action, value):
        """
            Show settings dialog
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        def do():
            from lollypop.dialog_settings import SettingsDialog
            dialog = SettingsDialog()
            dialog.show()
        # Settings dialog is heavy to load, let action close
        GLib.idle_add(do)

    def __on_about_activate(self, action, value):
        """
            Setup about dialog
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        def get_instance(children, instance):
            for child in children:
                if isinstance(child, instance):
                    return child
            return None
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/AboutDialog.ui")
        about = builder.get_object("about_dialog")
        about.set_transient_for(App().window)
        about.connect("response", self.__on_about_activate_response)
        about.show()

    def __on_shortcuts_activate(self, action, value):
        """
            Show shorctus
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/Shortcuts.ui")
        builder.get_object("shortcuts").set_transient_for(App().window)
        builder.get_object("shortcuts").show()

    def __on_status_changed(self, player, action):
        """
            Activate action if wanted
            @param player as Player
            @param action as Gio.SimpleAction
        """
        action.set_enabled(player.current_track.id is not None)

    def __on_seek_action(self, action, value):
        """
            Seek in stream
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        ms = value.get_int32()
        position = App().player.position
        seek = position + ms
        if seek < 0:
            seek = 0
        if seek > App().player.current_track.duration:
            seek = App().player.current_track.duration - 2
        App().player.seek(seek)

    def __on_player_action(self, action, value):
        """
            Change player state
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        string = value.get_string()
        if string == "play_pause":
            App().player.play_pause()
        elif string == "play":
            App().player.play()
        elif string == "stop":
            App().player.stop()
        elif string == "next":
            App().player.next()
        elif string == "next_album":
            App().player.skip_album()
        elif string == "prev":
            App().player.prev()
        elif string == "go_back":
            if App().window.toolbar.playback.back_button.get_sensitive():
                App().window.container.go_back()
        elif string == "reload":
            App().window.container.reload_view()
        elif string == "volume_up":
            App().player.set_volume(App().player.volume + 0.1)
        elif string == "volume_down":
            App().player.set_volume(App().player.volume - 0.1)
        elif string == "filter":
            App().window.container.show_filter()
        elif string == "loved":
            track = App().player.current_track
            if track.id is not None and track.id >= 0:
                loved = track.loved
                if loved & LovedFlags.LOVED:
                    loved &= ~ LovedFlags.LOVED
                else:
                    loved |= LovedFlags.LOVED
                track.set_loved(loved)
                if loved & LovedFlags.LOVED:
                    heart = "❤"
                else:
                    heart = "♡"
                App().notify.send("Lollypop", "%s - %s: %s" %
                                  (", ".join(track.artists),
                                   track.name,
                                   heart))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, Handy

from gettext import gettext as _

from lollypop.define import App, MARGIN


class Assistant(Handy.Window):
    """
        Dialog showing an assistant
        rules:
        [
            {
                "title": "plop" ,
                "icon_name": "folder",
                "markup": "sdmlqkdsqmkl",
                "uri_label": "flmdks",
                "uri": "http://....."
                "right_button_label": "Cancel",
                "right_button_style": "",
                "left_button_label": "Next",
                "left_button_style: "suggested-action"
            },
        ]
    """

    def __init__(self, rules):
        """
            Init assistant
            @param rules as {}
        """
        Handy.Window.__init__(self)
        self.__rules = rules
        self.__right_button = Gtk.Button.new()
        self.__right_button.show()
        self.__right_button.connect("clicked", self.__on_right_button_clicked)
        self.__left_button = Gtk.Button.new()
        self.__left_button.show()
        self.__left_button.connect("clicked", self.__on_left_button_clicked)
        self.__headerbar = Gtk.HeaderBar()
        self.__headerbar.show()
        self.__headerbar.pack_start(self.__right_button)
        self.__headerbar.pack_end(self.__left_button)
        self.__stack = Gtk.Stack.new()
        self.__stack.show()
        self.__stack.set_property("margin", MARGIN)
        box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        box.show()
        box.add(self.__headerbar)
        box.add(self.__stack)
        self.add(box)
        for rule in rules:
            page = self.__get_page(rule["icon_name"],
                                   rule["markup"],
                                   rule["uri_label"],
                                   rule["uri"])
            self.__stack.add_named(page, rule["title"])
        self.__set_page_visible(0)

    def update_uri(self, uri, index):
        """
            Update page URI at index
            @param uri as str
            @param index as int
        """
        self.__rules[index]["uri"] = uri

#######################
# PRIVATE             #
#######################
    def __set_page_visible(self, index):
        """
            Set page at index visible
            @param index as int
        """
        # Reset styles
        if index > 0:
            if self.__rules[index - 1]["left_button_style"] is not None:
                self.__left_button.get_style_context().remove_class(
                    self.__rules[index - 1]["left_button_style"])
            if self.__rules[index - 1]["right_button_style"] is not None:
                self.__left_button.get_style_context().remove_class(
                    self.__rules[index - 1]["right_button_style"])

        children = self.__stack.get_children()
        if index < len(children):
            self.__stack.set_visible_child(children[index])
            self.__headerbar.set_title(self.__rules[index]["title"])
            self.__left_button.set_label(
                self.__rules[index]["left_button_label"])
            if self.__rules[index]["right_button_label"] is None:
                self.__right_button.hide()
            else:
                self.__right_button.show()
                self.__right_button.set_label(
                    self.__rules[index]["right_button_label"])
            if self.__rules[index]["left_button_style"] is not None:
                self.__left_button.get_style_context().add_class(
                    self.__rules[index]["left_button_style"])
            if self.__rules[index]["right_button_style"] is not None:
                self.__left_button.get_style_context().add_class(
                    self.__rules[index]["right_button_style"])

    def __on_right_button_clicked(self, button):
        """
            Got back in stack
            @param button as Gtk.Button
        """
        visible = self.__stack.get_visible_child()
        index = self.__stack.get_children().index(visible)
        if index == 0:
            self.destroy()
        else:
            self.__set_page_visible(index - 1)

    def __on_left_button_clicked(self, button):
        """
            Got forward in stack
            @param button as Gtk.Button
        """
        visible = self.__stack.get_visible_child()
        index = self.__stack.get_children().index(visible)
        if index + 1 == len(self.__stack.get_children()):
            self.destroy()
        else:
            self.__set_page_visible(index + 1)

    def __on_uri_button_clicked(self, button):
        """
            Launch URI
            @param button as Gtk.Button
        """
        visible = self.__stack.get_visible_child()
        index = self.__stack.get_children().index(visible)
        uri = self.__rules[index]["uri"]
        if uri == "":
            button.set_label(_("Can't contact this service"))
            button.set_sensitive(False)
            self.__left_button.set_label(_("Finish"))
            self.__right_button.set_sensitive(False)
            for child in self.__stack.get_children()[index + 1:]:
                self.__stack.remove(child)
        Gtk.show_uri_on_window(App().window, uri, Gdk.CURRENT_TIME)

    def __get_page(self, icon_name, markup, uri_label, uri):
        """
            Get a new page
            @param icon_name as str
            @param text as str
            @param uri as str
        """
        grid = Gtk.Grid()
        grid.set_property("expand", True)
        grid.set_halign(Gtk.Align.CENTER)
        grid.set_row_spacing(20)
        grid.set_property("margin", 20)
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
        image.set_valign(Gtk.Align.START)
        label = Gtk.Label()
        label.set_markup(markup)
        grid.add(image)
        grid.add(label)
        if uri is not None:
            button = Gtk.Button.new_with_label(uri_label)
            button.connect("clicked", self.__on_uri_button_clicked)
            button.get_style_context().add_class("suggested-action")
            grid.add(button)
        grid.show_all()
        return grid



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, Gtk, GLib

from gettext import gettext as _

from lollypop.objects_album import Album
from lollypop.define import StorageType, MARGIN_SMALL, App
from lollypop.define import ViewType, Type
from lollypop.menu_playlists import PlaylistsMenu
from lollypop.menu_actions import ActionsMenu
from lollypop.menu_playback import TrackPlaybackMenu, AlbumPlaybackMenu
from lollypop.menu_sync import SyncAlbumsMenu
from lollypop.widgets_rating import RatingWidget
from lollypop.widgets_loved import LovedWidget


class AlbumMenu(Gio.Menu):
    """
        Contextual menu for album
    """

    def __init__(self, album, storage_type, view_type):
        """
            Init menu model
            @param album as Album
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        Gio.Menu.__init__(self)
        if App().window.folded:
            from lollypop.menu_header import AlbumMenuHeader
            self.append_item(AlbumMenuHeader(album))
        self.append_section(_("Playback"),
                            AlbumPlaybackMenu(album, view_type))
        if view_type & ViewType.ALBUM:
            from lollypop.menu_artist import ArtistAlbumsMenu
            menu = ArtistAlbumsMenu(album.artist_ids[0], album.storage_type)
            self.append_section(_("Artist"), menu)
        section = Gio.Menu()
        if album.storage_type & (StorageType.COLLECTION | StorageType.SAVED):
            section.append_submenu(_("Playlists"), PlaylistsMenu([album]))
        if album.storage_type & StorageType.COLLECTION:
            section.append_submenu(_("Devices"), SyncAlbumsMenu([album]))
        if section.get_n_items() != 0:
            self.append_section(_("Add to"), section)
        actions_menu = ActionsMenu(album)
        # Allow user to show album if not in album view
        if not view_type & ViewType.ALBUM:
            show_album_action = Gio.SimpleAction(name="show_album_action")
            App().add_action(show_album_action)
            show_album_action.connect(
                "activate",
                lambda x, y:
                App().window.container.show_view(
                   [Type.ALBUM], Album(album.id), storage_type))
            menu_item = Gio.MenuItem.new(_("Show album"),
                                         "app.show_album_action")
            menu_item.set_attribute_value("close", GLib.Variant("b", True))
            actions_menu.prepend_item(menu_item)
        if actions_menu.get_n_items() != 0:
            self.append_section(_("Others"), actions_menu)


class AlbumsMenu(Gio.Menu):
    """
        Contextual menu for albums
    """

    def __init__(self, title, albums, view_type):
        """
            Init menu model
            @param title as str
            @param albums as [Album]
            @param view_type as ViewType
        """
        Gio.Menu.__init__(self)
        if App().window.folded:
            from lollypop.menu_header import MenuHeader
            self.append_item(MenuHeader(title,
                                        "media-optical-cd-audio-symbolic"))
        self.append_section(_("Devices"), SyncAlbumsMenu(albums))


class TrackMenu(Gio.Menu):
    """
        Full Contextual menu for a track
    """

    def __init__(self, track, view_type):
        """
            Init menu model
            @param track as Track
            @param view_type as ViewType

        """
        Gio.Menu.__init__(self)
        if App().window.folded:
            from lollypop.menu_header import AlbumMenuHeader
            self.append_item(AlbumMenuHeader(track))
        self.append_section(_("Playback"), TrackPlaybackMenu(track, view_type))
        if view_type & ViewType.SEARCH:
            from lollypop.menu_artist import ArtistAlbumsMenu
            menu = ArtistAlbumsMenu(track.artist_ids[0], track.storage_type)
            if menu.get_n_items() != 0:
                self.append_section(_("Artist"), menu)
        if not track.storage_type & StorageType.EPHEMERAL:
            section = Gio.Menu()
            section.append_submenu(_("Playlists"), PlaylistsMenu([track]))
            self.append_section(_("Add to"), section)
        actions_menu = ActionsMenu(track)
        if actions_menu.get_n_items() != 0:
            self.append_section(_("Others"), actions_menu)


class TrackMenuExt(Gtk.Grid):
    """
        Additional widgets for track menu
    """

    def __init__(self, track):
        """
            Init widget
            @param track as Track
        """
        Gtk.Grid.__init__(self)
        self.set_margin_top(MARGIN_SMALL)
        self.set_row_spacing(MARGIN_SMALL)
        self.set_orientation(Gtk.Orientation.VERTICAL)

        if track.year is not None:
            year_label = Gtk.Label.new()
            year_label.set_text(str(track.year))
            dt = GLib.DateTime.new_from_unix_local(track.timestamp)
            year_label.set_tooltip_text(dt.format(_("%Y-%m-%d")))
            year_label.set_margin_end(5)
            year_label.get_style_context().add_class("dim-label")
            year_label.set_property("halign", Gtk.Align.START)
            year_label.set_property("hexpand", True)
            year_label.show()

        hgrid = Gtk.Grid()
        rating = RatingWidget(track)
        rating.set_property("halign", Gtk.Align.START)
        if App().window.folded:
            rating.set_icon_size(Gtk.IconSize.LARGE_TOOLBAR)
        rating.show()

        loved = LovedWidget(track)
        loved.set_property("halign", Gtk.Align.START)
        if App().window.folded:
            loved.set_icon_size(Gtk.IconSize.LARGE_TOOLBAR)
        loved.show()

        if track.year is not None:
            hgrid.add(year_label)
        hgrid.add(loved)
        hgrid.add(rating)
        hgrid.show()

        if track.is_web:
            from lollypop.helper_web import WebHelper
            helper = WebHelper(track, None)
            edit = Gtk.Entry()
            edit.set_placeholder_text(_("YouTube page address"))
            edit.set_margin_top(MARGIN_SMALL)
            edit.set_margin_start(MARGIN_SMALL)
            edit.set_margin_end(MARGIN_SMALL)
            edit.set_margin_bottom(MARGIN_SMALL)
            edit.set_property("hexpand", True)
            if helper.uri is not None:
                edit.set_text(helper.uri)
            edit.connect("changed", self.__on_edit_changed, track)
            edit.show()
            self.add(edit)

        separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        separator.show()
        self.add(separator)
        self.add(hgrid)

    @property
    def section(self):
        return None

    @property
    def submenu_name(self):
        return None

#######################
# PRIVATE             #
#######################
    def __on_edit_changed(self, edit, track):
        """
            Update track uri
            @param edit as Gtk.Edit
            @param track as Track
        """
        from urllib.parse import urlparse
        uri = edit.get_text()
        parsed = urlparse(uri)
        if parsed.scheme not in ["http", "https"]:
            return
        from lollypop.helper_web import WebHelper
        helper = WebHelper(track, None)
        helper.save(uri)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from lollypop.utils import popup_widget
from lollypop.view_lazyloading import LazyLoadingView
from lollypop.define import App, ViewType, MARGIN, StorageType
from lollypop.widgets_row_album import AlbumRow
from lollypop.widgets_listbox import ListBox
from lollypop.helper_gestures import GesturesHelper
from lollypop.helper_signals import SignalsHelper, signals_map


class AlbumsListView(LazyLoadingView, SignalsHelper, GesturesHelper):
    """
        View showing albums
    """

    @signals_map
    def __init__(self, genre_ids, artist_ids, view_type):
        """
            Init widget
            @param genre_ids as int
            @param artist_ids as int
            @param view_type as ViewType
        """
        LazyLoadingView.__init__(self, StorageType.ALL, view_type)
        self.__genre_ids = genre_ids
        self.__artist_ids = artist_ids
        self.__reveals = []
        # Calculate default album height based on current pango context
        # We may need to listen to screen changes
        self.__height = AlbumRow.get_best_height(self)
        self._box = ListBox()
        self._box.set_margin_bottom(MARGIN)
        self._box.set_margin_end(MARGIN)
        self._box.get_style_context().add_class("trackswidget")
        self._box.set_vexpand(True)
        self._box.set_selection_mode(Gtk.SelectionMode.NONE)
        self._box.show()
        GesturesHelper.__init__(self, self._box)
        if view_type & ViewType.DND:
            from lollypop.helper_dnd import DNDHelper
            self.__dnd_helper = DNDHelper(self._box, view_type)
            self.__dnd_helper.connect("dnd-insert", self.__on_dnd_insert)
        return [
            (App().player, "current-changed", "_on_current_changed"),
            (App().album_art, "album-artwork-changed", "_on_artwork_changed")
        ]

    def add_reveal_albums(self, albums):
        """
            Add albums to reveal list
            @param albums as [Album]
        """
        self.__reveals += list(albums)

    def add_value(self, album):
        """
            Insert item
            @param album as Album
        """
        # Merge album if previous is same
        if self.children and self.children[-1].album.id == album.id:
            track_ids = self.children[-1].album.track_ids
            for track in album.tracks:
                if track.id not in track_ids:
                    self.children[-1].tracks_view.append_row(track)
        else:
            LazyLoadingView.populate(self, [album])

    def insert_row(self, row, index):
        """
            Insert row at index
            @param row as AlbumRow
            @param index as int
        """
        row.connect("activated", self._on_row_activated)
        row.connect("album-removed", self._on_album_removed)
        row.connect("track-removed", self._on_track_removed)
        row.show()
        self._box.insert(row, index)

    def populate(self, albums):
        """
            Populate widget with album rows
            @param albums as [Album]
        """
        for child in self._box.get_children():
            self._box.remove(child)
        LazyLoadingView.populate(self, albums)

    def clear(self):
        """
            Clear the view
        """
        self.stop()
        self.__reveals = []
        for child in self._box.get_children():
            self._box.remove(child)

    def jump_to_current(self):
        """
            Scroll to album
        """
        y = self.__get_current_ordinate()
        if y is not None:
            self.scrolled.get_vadjustment().set_value(y)

    def destroy(self):
        """
            Force destroying the box
            Help freeing memory, no idea why
        """
        self._box.destroy()
        LazyLoadingView.destroy(self)

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"genre_ids": self.__genre_ids,
                "artist_ids": self.__artist_ids,
                "view_type": self.view_type & ~ViewType.SMALL}

    @property
    def dnd_helper(self):
        """
            Get Drag & Drop helper
            @return DNDHelper
        """
        return self.__dnd_helper

    @property
    def box(self):
        """
            Get album list box
            @return Gtk.ListBox
        """
        return self._box

    @property
    def children(self):
        """
            Get view children
            @return [AlbumRow]
        """
        return self._box.get_children()

#######################
# PROTECTED           #
#######################
    def _get_child(self, album):
        """
            Get an album view widget
            @param album as Album
            @return AlbumRow
        """
        if self.destroyed:
            return None
        row = AlbumRow(album, self.__height, self.view_type)
        row.connect("activated", self._on_row_activated)
        row.connect("album-removed", self._on_album_removed)
        row.connect("track-removed", self._on_track_removed)
        row.show()
        self._box.add(row)
        return row

    def _on_current_changed(self, player):
        """
            Update children state
            @param player as Player
        """
        for child in self._box.get_children():
            child.set_selection()

    def _on_artwork_changed(self, artwork, album_id):
        """
            Update children artwork if matching album id
            @param artwork as Artwork
            @param album_id as int
        """
        for child in self._box.get_children():
            if child.album.id == album_id:
                child.set_artwork()

    def _on_primary_long_press_gesture(self, x, y):
        """
            Show row menu
            @param x as int
            @param y as int
        """
        self.__popup_menu(x, y)

    def _on_primary_press_gesture(self, x, y, event):
        """
            Activate current row
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        row = self._box.get_row_at_y(y)
        if row is None:
            return
        self._box.set_selection_mode(Gtk.SelectionMode.NONE)
        row.reveal()

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Show row menu
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        self._on_primary_long_press_gesture(x, y)

    def _on_populated(self, widget):
        """
            Add another album/disc
            @param widget as AlbumWidget/TracksView
        """
        if widget.album in self.__reveals:
            widget.reveal()
            self.__reveals.remove(widget.album)
        else:
            LazyLoadingView._on_populated(self, widget)

    def _on_row_activated(self, row, track):
        pass

    def _on_album_removed(self, row):
        pass

    def _on_track_removed(self, row, track):
        pass

#######################
# PRIVATE             #
#######################
    def __get_current_ordinate(self):
        """
            If current track in widget, return it ordinate,
            @return y as int
        """
        y = None
        for child in self._box.get_children():
            if child.album == App().player.current_track.album:
                child.reveal(True)
                y = child.translate_coordinates(self._box, 0, 0)[1]
        return y

    def __popup_menu(self, x, y):
        """
            Popup menu for album
            @param x as int
            @param y as int
        """
        row = self._box.get_row_at_y(y)
        if row is None:
            return
        # First check it's not a track gesture
        if row.revealed:
            for track_row in row.listbox.get_children():
                coordinates = track_row.translate_coordinates(self._box, 0, 0)
                if coordinates is not None and\
                        coordinates[1] < y and\
                        coordinates[1] + track_row.get_allocated_height() > y:
                    track_row.popup_menu(self._box, x, y)
                    return
        # Then handle album gesture
        from lollypop.menu_objects import AlbumMenu
        from lollypop.widgets_menu import MenuBuilder
        menu = AlbumMenu(row.album, ViewType.ALBUM, self.view_type)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, self._box, x, y, row)

    def __reveal_row(self, row):
        """
            Reveal row if style always present
        """
        style_context = row.get_style_context()
        if style_context.has_class("drag-down"):
            row.reveal(True)

    def __on_dnd_insert(self, dnd_helper, row, index):
        """
            Insert row at index
            @param dnd_helper as DNDHelper
            @param row as AlbumRow
            @param index as int
        """
        self.insert_row(row, index)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.widgets_typeahead import TypeAheadWidget
from lollypop.define import App


class FilterContainer:
    """
        Filtering management
    """

    def __init__(self):
        """
            Init container
        """
        self.__index = 0
        self.__type_ahead = TypeAheadWidget()
        self.__type_ahead.show()
        self.grid_view.add(self.__type_ahead)

    def show_filter(self):
        """
            Show filtering widget
        """
        reveal = not self.__type_ahead.get_reveal_child()
        if reveal:
            self.__type_ahead.set_reveal_child(True)
            App().enable_special_shortcuts(False)
            self.__type_ahead.entry.grab_focus()
        elif self.__type_ahead.entry.has_focus():
            self.__type_ahead.set_reveal_child(False)
            App().enable_special_shortcuts(True)
            self.__type_ahead.entry.set_text("")
        else:
            self.__type_ahead.entry.grab_focus()

    @property
    def type_ahead(self):
        """
            Get typeahead widget
            @return TypeAheadWidget
        """
        return self.__type_ahead

############
# PRIVATE  #
############



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, GObject, Pango

from lollypop.define import ArtSize, ViewType, MARGIN, Size
from lollypop.widgets_banner import BannerWidget
from lollypop.utils import emit_signal, get_title_for_genres_artists


class FlowboxBannerWidget(BannerWidget):
    """
        Banner for flowbox
    """

    __gsignals__ = {
        "play-all": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "show-menu": (GObject.SignalFlags.RUN_FIRST, None,
                      (GObject.TYPE_PYOBJECT,))
    }

    def __init__(self, genre_ids, artist_ids, view_type, show_menu=False):
        """
            Init banner
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param view_type as ViewType
            @param show_menu as bool
        """
        BannerWidget.__init__(self, view_type | ViewType.OVERLAY)
        grid = Gtk.Grid.new()
        grid.show()
        grid.set_property("valign", Gtk.Align.CENTER)
        self.__title_label = Gtk.Label.new()
        self.__title_label.show()
        self.__title_label.set_margin_start(MARGIN)
        self.__title_label.set_hexpand(True)
        self.__title_label.set_property("halign", Gtk.Align.START)
        self.__title_label.set_ellipsize(Pango.EllipsizeMode.END)
        linked = Gtk.Grid.new()
        linked.show()
        linked.get_style_context().add_class("linked")
        linked.set_margin_end(MARGIN)
        linked.set_property("halign", Gtk.Align.END)
        self.__play_button = Gtk.Button.new_from_icon_name(
            "media-playback-start-symbolic", Gtk.IconSize.BUTTON)
        self.__play_button.show()
        self.__play_button.get_style_context().add_class("banner-button")
        self.__play_button.connect("clicked", self.__on_play_button_clicked)
        self.__shuffle_button = Gtk.Button.new_from_icon_name(
            "media-playlist-shuffle-symbolic", Gtk.IconSize.BUTTON)
        self.__shuffle_button.show()
        self.__shuffle_button.get_style_context().add_class("banner-button")
        self.__shuffle_button.connect("clicked",
                                      self.__on_shuffle_button_clicked)
        linked.add(self.__play_button)
        linked.add(self.__shuffle_button)
        if show_menu:
            self.__menu_button = Gtk.Button.new_from_icon_name(
                "view-more-symbolic", Gtk.IconSize.BUTTON)
            self.__menu_button.show()
            self.__menu_button.get_style_context().add_class("banner-button")
            self.__menu_button.connect("clicked",
                                       self.__on_menu_button_clicked)
            linked.add(self.__menu_button)
        grid.add(self.__title_label)
        grid.add(linked)
        self._overlay.add_overlay(grid)
        self._overlay.set_overlay_pass_through(grid, True)
        title_str = get_title_for_genres_artists(genre_ids, artist_ids)
        self.__title_label.set_markup("<b>%s</b>" %
                                      GLib.markup_escape_text(title_str))

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__set_internal_size()

    @property
    def height(self):
        """
            Get wanted height
            @return int
        """
        return ArtSize.SMALL

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__set_internal_size()

#######################
# PRIVATE             #
#######################
    def __set_internal_size(self):
        """
            Set content size based on current width
        """
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        if self.width <= Size.MEDIUM:
            self.__title_label.get_style_context().add_class(
                "text-large")
        else:
            self.__title_label.get_style_context().add_class(
                "text-x-large")

    def __on_play_button_clicked(self, button):
        """
            Play playlist
            @param button as Gtk.Button
        """
        emit_signal(self, "play-all", False)

    def __on_shuffle_button_clicked(self, button):
        """
            Play playlist shuffled
            @param button as Gtk.Button
        """
        emit_signal(self, "play-all", True)

    def __on_menu_button_clicked(self, button):
        """
            Show sync menu
            @param button as Gtk.Button
        """
        self.emit("show-menu", button)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Pango

from gettext import gettext as _

from lollypop.widgets_row_track import TrackRow
from lollypop.widgets_label import LabelWidget
from lollypop.objects_album import Album
from lollypop.utils import emit_signal
from lollypop.define import App, ViewType, LovedFlags
from lollypop.view_tracks import TracksView


class AlbumTracksView(TracksView):
    """
        Responsive view showing an album tracks
    """

    def __init__(self, album, view_type):
        """
            Init view
            @param album as Album
            @param view_type as ViewType
        """
        TracksView.__init__(self, view_type)
        self.__album = album
        self.__discs_to_load = []
        self.__disc_labels = {}
        self.__show_disc_header = False
        self.__populated = False
        self.__show_tag_tracknumber = App().settings.get_value(
            "show-tag-tracknumber")

    def populate(self):
        """
            Populate tracks lazy
        """
        self._init()
        if self.__discs_to_load:
            disc = self.__discs_to_load.pop(0)
            disc_number = disc.number
            tracks = disc.tracks
            items = []
            if self.view_type & ViewType.SINGLE_COLUMN:
                items.append((self._tracks_widget_left[0], tracks))
            else:
                mid_tracks = int(0.5 + len(tracks) / 2)
                items.append((self._tracks_widget_left[disc_number],
                              tracks[:mid_tracks]))
                items.append((self._tracks_widget_right[disc_number],
                              tracks[mid_tracks:]))
            self.__set_disc_header(disc_number)
            self.__load_disc(items, disc_number)
        else:
            self.__populated = True
            emit_signal(self, "populated")
            if not self.children:
                label = Gtk.Label.new(_("All tracks skipped"))
                label.show()
                self._responsive_widget.insert_row(0)
                self._responsive_widget.attach(label, 0, 0, 2, 1)

    def append_row(self, track):
        """
            Append a track
            ONE COLUMN ONLY
            @param track as Track
            @param position as int
        """
        self._init()
        if not self.is_populated:
            self.populate()
        self.__album.append_track(track)
        for key in self._tracks_widget_left.keys():
            self._add_tracks(self._tracks_widget_left[key], [track])
            return

    def append_rows(self, tracks):
        """
            Add track rows
            ONE COLUMN ONLY
            @param tracks as [Track]
        """
        self._init()
        if not self.is_populated:
            self.populate()
        self.__album.append_tracks(tracks)
        for key in self._tracks_widget_left.keys():
            self._add_tracks(self._tracks_widget_left[key], tracks)
            return

    def remove_row(self, track):
        """
            Remove row for track
            @param track as Track
        """
        self.__album.remove_track(track)
        for child in self.children:
            if child.track == track:
                child.destroy()

    @property
    def children(self):
        """
            Return all rows
            @return [Gtk.ListBoxRow]
        """
        self._init()
        rows = []
        for disc in self.__album.discs:
            if disc.number in self._tracks_widget_left.keys():
                widget = self._tracks_widget_left[disc.number]
                rows += widget.get_children()
            if disc.number in self._tracks_widget_right.keys():
                widget = self._tracks_widget_right[disc.number]
                rows += widget.get_children()
        return rows

    @property
    def boxes(self):
        """
            @return [Gtk.ListBox]
        """
        boxes = []
        for widget in self._tracks_widget_left.values():
            boxes.append(widget)
        for widget in self._tracks_widget_right.values():
            boxes.append(widget)
        return boxes

    @property
    def discs(self):
        """
            Get widget discs
            @return [Discs]
        """
        return self.__album.discs

    @property
    def is_populated(self):
        """
            Return True if populated
            @return bool
        """
        return self.__populated

#######################
# PROTECTED           #
#######################
    def _init(self):
        """
            Init main widget
        """
        if TracksView._init(self):
            if self.view_type & ViewType.SINGLE_COLUMN:
                self.__album.merge_discs()
            for disc in self.__album.discs:
                self._add_disc_container(disc.number)
            self.__discs_to_load = list(self.__album.discs)
            self.__show_disc_header = len(self.__discs_to_load) > 1

    def _add_tracks(self, widget, tracks, position=0):
        """
            Add tracks to widget
            @param widget as Gtk.ListBox
            @param tracks as [Track]
        """
        for track in tracks:
            # If user does not want to show real tracknumber and we are
            # in album view, calculate a fake tracknumber
            if not self.__show_tag_tracknumber and\
                    self.view_type & (ViewType.ALBUM | ViewType.ARTIST):
                track.set_number(position + 1)
            row = TrackRow(track, self.__album.artist_ids, self.view_type)
            row.show()
            row.connect("removed", self.__on_track_row_removed)
            widget.add(row)
            position += 1

    def _set_orientation(self, orientation):
        """
            Set columns orientation
            @param orientation as Gtk.Orientation
        """
        if not TracksView._set_orientation(self, orientation):
            return
        idx = 0
        # Vertical
        ##########################
        #  --------Label-------- #
        #  |     Column 1      | #
        #  |     Column 2      | #
        ##########################
        # Horizontal
        ###########################
        # ---------Label--------- #
        # | Column 1 | Column 2 | #
        ###########################
        for disc in self.__album.discs:
            if disc.number not in self._tracks_widget_left.keys():
                continue
            label = self.__get_disc_label(disc.number)
            if orientation == Gtk.Orientation.VERTICAL:
                self._responsive_widget.attach(
                    label, 0, idx, 1, 1)
            else:
                self._responsive_widget.attach(
                    label, 0, idx, 2, 1)
            idx += 1
            if orientation == Gtk.Orientation.VERTICAL:
                self._responsive_widget.attach(
                          self._tracks_widget_left[disc.number],
                          0, idx, 2, 1)
                idx += 1
            else:
                self._responsive_widget.attach(
                          self._tracks_widget_left[disc.number],
                          0, idx, 1, 1)
            if not self.view_type & ViewType.SINGLE_COLUMN:
                if orientation == Gtk.Orientation.VERTICAL:
                    self._responsive_widget.attach(
                               self._tracks_widget_right[disc.number],
                               0, idx, 2, 1)
                else:
                    self._responsive_widget.attach(
                               self._tracks_widget_right[disc.number],
                               1, idx, 1, 1)
            idx += 1

    def _on_loading_changed(self, player, status, track):
        """
            Update row loading status
            @param player as Player
            @param status as bool
            @param track as Track
        """
        for child in self.children:
            child.get_parent().unselect_row(child)
        self._disallow_selection()
        if self.__album.is_web:
            TracksView._on_loading_changed(self, player, status, track)

    def _on_activated(self, widget, track):
        """
            Handle playback if album or pass signal
            @param widget as TracksWidget
            @param track as Track
        """
        if self.view_type & (ViewType.ALBUM | ViewType.ARTIST):
            tracks = []
            for child in self.children:
                if not child.track.loved & LovedFlags.SKIPPED or\
                        track.id == child.track.id:
                    tracks.append(child.track)
                child.set_state_flags(Gtk.StateFlags.NORMAL, True)
            if not App().player.is_party:
                album = Album(track.album.id, [], [])
                album.set_tracks(tracks)
                if not App().settings.get_value("append-albums"):
                    App().player.clear_albums()
                App().player.add_album(album)
                App().player.load(album.get_track(track.id))
            else:
                App().player.load(track)
        else:
            emit_signal(self, "activated", track)

    def _on_row_selected(self, listbox, row):
        """
            Update album tracks
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        TracksView._on_row_selected(self, listbox, row)
        selected_rows = []
        for child in self.children:
            if child.is_selected():
                selected_rows.append(child)
        if selected_rows:
            tracks = [row.track for row in selected_rows]
            self.__album.set_tracks(tracks)
        else:
            self.__album.reset_tracks()

#######################
# PRIVATE             #
#######################
    def __get_disc_label(self, disc_number):
        """
            Get disc label widget
            @param disc_number as int
            @return Gtk.Label
        """
        if disc_number in self.__disc_labels.keys():
            label = self.__disc_labels[disc_number]
        else:
            label = LabelWidget()
            label.set_ellipsize(Pango.EllipsizeMode.END)
            label.set_property("halign", Gtk.Align.START)
            label.get_style_context().add_class("dim-label")
            label.connect("clicked", self.__on_label_clicked, disc_number)
            label.set_tooltip_text(_("Play"))
            self.__disc_labels[disc_number] = label
        return label

    def __set_disc_header(self, disc_number):
        """
            Set disc header
            @param disc_number as int
        """
        if self.__show_disc_header:
            label = self.__get_disc_label(disc_number)
            disc_names = self.__album.disc_names(disc_number)
            if disc_names:
                label_str = ", ".join(disc_names)
            else:
                label_str = _("Disc %s") % disc_number
            disc_year = App().tracks.get_year_for_album(
                self.__album.id, disc_number)
            if disc_year != self.__album.year:
                markup = "%s  <span size='small' alpha='50000'>(%s)</span>"
                markup = markup % (GLib.markup_escape_text(label_str),
                                   disc_year)
            else:
                markup = GLib.markup_escape_text(label_str)
            label.set_markup(markup)
            label.show()

    def __load_disc(self, items, disc_number, position=0):
        """
            Load discs
            @param items as (TrackWidget, [Tracks])
            @param disc_number as int
            @param position as int
        """
        if items:
            (widget, tracks) = items.pop(0)
            self._add_tracks(widget, tracks, position)
            position += len(tracks)
            widget.show()
            GLib.idle_add(self.__load_disc, items, disc_number, position)
        else:
            emit_signal(self, "populated")

    def __on_track_row_removed(self, row):
        """
            Pass signal
            @param row as TrackRow
        """
        emit_signal(self, "track-removed", row)

    def __on_label_clicked(self, label, disc_number):
        """
            Add disc to playback
            @param label as LabelWidget
            @param disc_number as int
        """
        album = self.__album.clone(False)
        album.set_disc_number(disc_number)
        App().player.play_album(album)
